import os
import shutil
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from flask import Blueprint, current_app, request, send_file
from sqlalchemy import and_, asc, desc, exc, func, nullslast, or_

from fittrackee import appLog, db, limiter
from fittrackee.dates import get_readable_duration
from fittrackee.emails.tasks import send_email
from fittrackee.equipments.models import Equipment
from fittrackee.files import get_absolute_file_path
from fittrackee.oauth2.server import require_auth
from fittrackee.reports.models import ReportAction
from fittrackee.responses import (
    ForbiddenErrorResponse,
    HttpResponse,
    InvalidPayloadErrorResponse,
    NotFoundErrorResponse,
    UserNotFoundErrorResponse,
    handle_error_and_return_response,
)
from fittrackee.visibility_levels import VisibilityLevel
from fittrackee.workouts.models import Record, Workout, WorkoutSegment

from .exceptions import (
    BlockUserException,
    FollowRequestAlreadyRejectedError,
    InvalidEmailException,
    InvalidUserRole,
    NotExistingFollowRequestError,
    OwnerException,
    UserNotFoundException,
)
from .models import FollowRequest, User, UserDataExport, UserSportPreference
from .roles import UserRole
from .users_service import UserManagerService
from .utils.language import get_language

if TYPE_CHECKING:
    from sqlalchemy.sql.expression import (
        BinaryExpression,
        ColumnElement,
        UnaryExpression,
    )

users_blueprint = Blueprint("users", __name__)

ACTIONS_PER_PAGE = 5
USERS_PER_PAGE = 10
EMPTY_USERS_RESPONSE = {
    "status": "success",
    "data": {"users": []},
    "pagination": {
        "has_next": False,
        "has_prev": False,
        "page": 1,
        "pages": 0,
        "total": 0,
    },
}
WORKOUTS_PER_PAGE = 5


def _get_value_depending_on_user_rights(
    params: Dict, key: str, auth_user: Optional[User]
) -> str:
    value = params.get(key, "false").lower()
    if not auth_user or not auth_user.has_admin_rights:
        value = "false"
    return value


def get_users_list(auth_user: User) -> Dict:
    params = request.args.copy()

    query = params.get("q")
    page = int(params.get("page", 1))
    per_page = int(params.get("per_page", USERS_PER_PAGE))
    if per_page > 50:
        per_page = 50
    column = params.get("order_by", "username")
    user_column = getattr(User, column)
    order = params.get("order", "asc")
    order_clauses: List["UnaryExpression"] = [
        asc(user_column) if order == "asc" else desc(user_column)
    ]
    if column != "username":
        order_clauses.append(User.username.asc())
    if column == "suspended_at":
        order_clauses = [nullslast(order_clauses[0])]
    with_inactive = _get_value_depending_on_user_rights(
        params, "with_inactive", auth_user
    )
    with_hidden_users = _get_value_depending_on_user_rights(
        params, "with_hidden", auth_user
    )
    with_suspended_users = _get_value_depending_on_user_rights(
        params, "with_suspended", auth_user
    )
    with_following = params.get("with_following", "false").lower()
    following_user_ids = (
        auth_user.get_following_user_ids() if with_following == "true" else []
    )

    filters: List[Union["ColumnElement", "BinaryExpression"]] = []
    if query:
        filters.append(User.username.ilike("%" + query + "%"))
    if with_inactive != "true":
        filters.append(User.is_active == True)  # noqa
    if with_hidden_users != "true":
        filters.append(
            (
                or_(
                    User.hide_profile_in_users_directory == False,  # noqa
                    and_(
                        User.id.in_(following_user_ids),
                        User.hide_profile_in_users_directory == True,  # noqa
                    ),
                )
            ),
        )
    if with_suspended_users != "true":
        filters.append(User.suspended_at == None)  # noqa
    users_pagination = (
        User.query.filter(*filters)
        .order_by(*order_clauses)
        .paginate(page=page, per_page=per_page, error_out=False)
    )
    users = users_pagination.items
    return {
        "status": "success",
        "data": {
            "users": [user.serialize(current_user=auth_user) for user in users]
        },
        "pagination": {
            "has_next": users_pagination.has_next,
            "has_prev": users_pagination.has_prev,
            "page": users_pagination.page,
            "pages": users_pagination.pages,
            "total": users_pagination.total,
        },
    }


@users_blueprint.route("/users", methods=["GET"])
@require_auth(scopes=["users:read"])
def get_users(auth_user: User) -> Dict:
    """
    Get all users.
    If authenticated user has admin rights, users email is returned.

    It returns user preferences only for authenticated user.

    **Scope**: ``users:read``

    **Example request**:

    - without parameters:

    .. sourcecode:: http

      GET /api/users HTTP/1.1
      Content-Type: application/json

    - with some query parameters:

    .. sourcecode:: http

      GET /api/users?order_by=workouts_count&par_page=5  HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "users": [
            {
              "admin": true,
              "bio": null,
              "birth_date": null,
              "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
              "email": "admin@example.com",
              "first_name": null,
              "followers": 0,
              "following": 0,
              "follows": "false",
              "is_followed_by": "false",
              "last_name": null,
              "location": null,
              "map_visibility": "private",
              "nb_sports": 3,
              "nb_workouts": 6,
              "picture": false,
              "records": [
                {
                  "id": 9,
                  "record_type": "AS",
                  "sport_id": 1,
                  "user": "admin",
                  "value": 18,
                  "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                  "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                },
                {
                  "id": 10,
                  "record_type": "FD",
                  "sport_id": 1,
                  "user": "admin",
                  "value": 18,
                  "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                  "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                },
                {
                  "id": 13,
                  "record_type": "HA",
                  "sport_id": 1,
                  "user": "Sam",
                  "value": 43.97,
                  "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                  "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                },
                {
                  "id": 11,
                  "record_type": "LD",
                  "sport_id": 1,
                  "user": "admin",
                  "value": "1:01:00",
                  "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                  "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                },
                {
                  "id": 12,
                  "record_type": "MS",
                  "sport_id": 1,
                  "user": "admin",
                  "value": 18,
                  "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                  "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                }
              ],
              "sports_list": [
                  1,
                  4,
                  6
              ],
              "total_distance": 67.895,
              "total_duration": "6:50:27",
              "username": "admin",
              "workouts_visibility": "private"
            },
            {
              "admin": false,
              "bio": null,
              "birth_date": null,
              "created_at": "Sat, 20 Jul 2019 11:27:03 GMT",
              "email": "sam@example.com",
              "first_name": null,
              "followers": 0,
              "following": 0,
              "follows": "false",
              "is_followed_by": "false",
              "last_name": null,
              "location": null,
              "map_visibility": "private",
              "nb_sports": 0,
              "nb_workouts": 0,
              "picture": false,
              "records": [],
              "sports_list": [],
              "total_distance": 0,
              "total_duration": "0:00:00",
              "username": "sam",
              "workouts_visibility": "private"
            }
          ]
        },
        "status": "success"
      }

    :query integer page: page if using pagination (default: 1)
    :query integer per_page: number of users per page (default: 10, max: 50)
    :query string q: query on user name
    :query string order: sorting order: ``asc``, ``desc`` (default: ``asc``)
    :query string order_by: sorting criteria: ``username``, ``created_at``,
                            ``workouts_count``, ``role``, ``is_active``
                            (default: ``username``)
    :query boolean with_following: returns hidden users followed by user if
           true
    :query boolean with_hidden_users: returns hidden users if ``true`` (only if
           authenticated user has administration rights - for users
           administration)
    :query boolean with_inactive: returns inactive users if ``true`` (only if
           authenticated user has administration rights - for users
           administration)
    :query boolean with_suspended: returns suspended users if ``true`` (only if
           authenticated user has administration rights - for users
           administration)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``

    """
    return get_users_list(auth_user)


@users_blueprint.route("/users/<user_name>", methods=["GET"])
@require_auth(scopes=["users:read"], optional_auth_user=True)
def get_single_user(
    auth_user: Optional[User], user_name: str
) -> Union[Dict, HttpResponse]:
    """
    Get single user details.
    If a user is authenticated, it returns relationships.
    If authenticated user has admin rights, user email is returned.

    It returns user preferences only for authenticated user.

    **Scope**: ``users:read`` for Oauth 2.0 client

    **Example request**:

    .. sourcecode:: http

      GET /api/users/admin HTTP/1.1
      Content-Type: application/json

    **Example response**:

    - when a user is authenticated:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": [
          {
            "admin": true,
            "bio": null,
            "birth_date": null,
            "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
            "email": "admin@example.com",
            "first_name": null,
            "followers": 0,
            "following": 0,
            "follows": "false",
            "is_followed_by": "false",
            "last_name": null,
            "location": null,
            "map_visibility": "private",
            "nb_sports": 3,
            "nb_workouts": 6,
            "picture": false,
            "records": [
              {
                "id": 9,
                "record_type": "AS",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 10,
                "record_type": "FD",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 13,
                "record_type": "HA",
                "sport_id": 1,
                "user": "Sam",
                "value": 43.97,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 11,
                "record_type": "LD",
                "sport_id": 1,
                "user": "admin",
                "value": "1:01:00",
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 12,
                "record_type": "MS",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              }
            ],
            "sports_list": [
                1,
                4,
                6
            ],
            "total_distance": 67.895,
            "total_duration": "6:50:27",
            "username": "admin",
            "workouts_visibility": "private"
          }
        ],
        "status": "success"
      }

    - when no authentication:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": [
          {
            "admin": true,
            "bio": null,
            "birth_date": null,
            "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
            "email": "admin@example.com",
            "first_name": null,
            "followers": 0,
            "following": 0,
            "follows": "false",
            "is_followed_by": "false",
            "last_name": null,
            "location": null,
            "map_visibility": "private",
            "nb_workouts": 6,
            "picture": false,
            "username": "admin",
            "workouts_visibility": "private"
          }
        ],
        "status": "success"
      }

    :param integer user_name: user name

    :reqheader Authorization: OAuth 2.0 Bearer Token if user is authenticated

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user does not exist``
    """
    try:
        user = User.query.filter(
            func.lower(User.username) == func.lower(user_name),
        ).first()
        if user:
            if (
                not auth_user or not auth_user.has_admin_rights
            ) and not user.is_active:
                return UserNotFoundErrorResponse()
            return {
                "status": "success",
                "data": {
                    "users": [
                        user.serialize(current_user=auth_user, light=False)
                    ]
                },
            }
    except (ValueError, UserNotFoundException):
        pass
    return UserNotFoundErrorResponse()


@users_blueprint.route("/users/<user_name>/picture", methods=["GET"])
@limiter.exempt
def get_picture(user_name: str) -> Any:
    """get user picture

    **Example request**:

    .. sourcecode:: http

      GET /api/users/admin/picture HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: image/jpeg

    :param integer user_name: user name

    :statuscode 200: ``success``
    :statuscode 404:
        - ``user does not exist``
        - ``No picture.``

    """
    try:
        user = User.query.filter(
            func.lower(User.username) == func.lower(user_name),
        ).first()
        if not user:
            return UserNotFoundErrorResponse()
        if user.picture is not None:
            picture_path = get_absolute_file_path(user.picture)
            return send_file(picture_path)
    except UserNotFoundException:
        return UserNotFoundErrorResponse()
    except Exception:  # nosec
        pass
    return NotFoundErrorResponse("No picture.")


@users_blueprint.route("/users/<user_name>", methods=["PATCH"])
@require_auth(scopes=["users:write"], role=UserRole.ADMIN)
def update_user(auth_user: User, user_name: str) -> Union[Dict, HttpResponse]:
    """
    Update user account.

    - add/remove admin rights (regardless user account status)
    - reset password (and send email to update user password,
      if sending enabled)
    - update user email (and send email to new user email, if sending enabled)
    - activate account for an inactive user
    - deactivate account after report.

    **Scope**: ``users:write``

    **Minimum role**: Administrator

    **Example request**:

    .. sourcecode:: http

      PATCH /api/users/<user_name> HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": [
          {
            "admin": true,
            "bio": null,
            "birth_date": null,
            "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
            "email": "admin@example.com",
            "first_name": null,
            "followers": 0,
            "following": 0,
            "follows": "false",
            "is_followed_by": "false",
            "last_name": null,
            "location": null,
            "map_visibility": "private",
            "nb_workouts": 6,
            "nb_sports": 3,
            "picture": false,
            "records": [
              {
                "id": 9,
                "record_type": "AS",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 10,
                "record_type": "FD",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 13,
                "record_type": "HA",
                "sport_id": 1,
                "user": "Sam",
                "value": 43.97,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 11,
                "record_type": "LD",
                "sport_id": 1,
                "user": "admin",
                "value": "1:01:00",
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 12,
                "record_type": "MS",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              }
            ],
            "sports_list": [
                1,
                4,
                6
            ],
            "total_distance": 67.895,
            "total_duration": "6:50:27",
            "username": "admin",
            "workouts_visibility": "private"
          }
        ],
        "status": "success"
      }

    :param string user_name: user name

    :<json boolean activate: (de-)activate user account
    :<json boolean role: user role (``user``, ``admin``, ``moderator``).
                   ``owner`` can only be set via **CLI**.
    :<json boolean new_email: new user email
    :<json boolean reset_password: reset user password

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid payload``
        - ``invalid role``
        - ``valid email must be provided``
        - ``new email must be different than current email``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user does not exist``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    user_data = request.get_json()
    if not user_data:
        return InvalidPayloadErrorResponse()

    activate = user_data.get("activate")
    if activate is False and user_name == auth_user.username:
        return ForbiddenErrorResponse()

    role = user_data.get("role")
    if role == "owner":
        return InvalidPayloadErrorResponse(
            "'owner' can not be set via API, please user CLI instead"
        )

    try:
        reset_password = user_data.get("reset_password", False)
        new_email = user_data.get("new_email")
        user_manager_service = UserManagerService(
            username=user_name, moderator_id=auth_user.id
        )
        user, _, _, _ = user_manager_service.update(
            role=role,
            activate=user_data.get("activate"),
            reset_password=reset_password,
            new_email=new_email,
            with_confirmation=current_app.config["CAN_SEND_EMAILS"],
            raise_error_on_owner=True,
        )

        if current_app.config["CAN_SEND_EMAILS"]:
            user_language = get_language(user.language)
            fittrackee_url = current_app.config["UI_URL"]
            if reset_password:
                user_data = {
                    "language": user_language,
                    "email": user.email,
                }
                send_email.send(
                    user_data,
                    email_data={
                        "username": user.username,
                        "fittrackee_url": fittrackee_url,
                    },
                    template="password_change",
                )
                password_reset_token = user.encode_password_reset_token(
                    user.id
                )
                send_email.send(
                    user_data,
                    email_data={
                        "expiration_delay": get_readable_duration(
                            current_app.config[
                                "PASSWORD_TOKEN_EXPIRATION_SECONDS"
                            ],
                            user_language,
                        ),
                        "username": user.username,
                        "password_reset_url": (
                            f"{fittrackee_url}/password-reset?"
                            f"token={password_reset_token}"
                        ),
                        "fittrackee_url": fittrackee_url,
                    },
                    template="password_reset_request",
                )

            if new_email:
                user_data = {
                    "language": user_language,
                    "email": user.email_to_confirm,
                }
                email_data = {
                    "username": user.username,
                    "fittrackee_url": fittrackee_url,
                    "email_confirmation_url": (
                        f"{fittrackee_url}/email-update"
                        f"?token={user.confirmation_token}"
                    ),
                }
                send_email.send(
                    user_data, email_data, template="email_update_to_new_email"
                )

        return {
            "status": "success",
            "data": {
                "users": [user.serialize(current_user=auth_user, light=False)]
            },
        }
    except UserNotFoundException:
        return UserNotFoundErrorResponse()
    except (InvalidEmailException, InvalidUserRole, OwnerException) as e:
        return InvalidPayloadErrorResponse(str(e))
    except (TypeError, exc.StatementError) as e:
        return handle_error_and_return_response(e, db=db)


@users_blueprint.route("/users/<user_name>", methods=["DELETE"])
@require_auth(scopes=["users:write"], allow_suspended_user=True)
def delete_user(
    auth_user: User, user_name: str
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Delete a user account.

    A user can only delete his own account.

    A user with admin rights can delete all accounts except his account if
    he is the only user with admin rights.
    Only owner can delete his own account.

    Suspended user can access this endpoint.

    **Scope**: ``users:write``

    **Example request**:

    .. sourcecode:: http

      DELETE /api/users/john_doe HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 204 NO CONTENT
      Content-Type: application/json

    :param string user_name: user name

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 204: user account deleted
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you can not delete your account, no other user has admin rights``
    :statuscode 404:
        - ``user does not exist``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    try:
        user = User.query.filter(
            func.lower(User.username) == func.lower(user_name),
        ).first()
        if not user:
            return UserNotFoundErrorResponse()
        if user.id != auth_user.id and user.role == UserRole.OWNER.value:
            return ForbiddenErrorResponse("you can not delete owner account")

        if user.id != auth_user.id and not auth_user.has_admin_rights:
            return ForbiddenErrorResponse()
        if (
            user.has_admin_rights is True
            and User.query.filter(User.role >= UserRole.ADMIN.value).count()
            == 1
        ):
            return ForbiddenErrorResponse(
                "you can not delete your account, "
                "no other user has admin rights"
            )

        db.session.query(UserSportPreference).filter(
            UserSportPreference.user_id == user.id
        ).delete()
        db.session.query(Record).filter(Record.user_id == user.id).delete()
        # delete all equipment associated with this user
        db.session.query(Equipment).filter(
            Equipment.user_id == user.id
        ).delete()
        db.session.query(WorkoutSegment).filter(
            WorkoutSegment.workout_id == Workout.id, Workout.user_id == user.id
        ).delete(synchronize_session=False)
        db.session.query(Workout).filter(Workout.user_id == user.id).delete()
        db.session.query(UserDataExport).filter(
            UserDataExport.user_id == user.id
        ).delete()
        db.session.flush()
        user_picture = user.picture
        db.session.delete(user)
        db.session.commit()
        if user_picture:
            picture_path = get_absolute_file_path(user.picture)
            if os.path.isfile(picture_path):
                os.remove(picture_path)
        shutil.rmtree(
            get_absolute_file_path(f"exports/{user.id}"),
            ignore_errors=True,
        )
        shutil.rmtree(
            get_absolute_file_path(f"workouts/{user.id}"),
            ignore_errors=True,
        )
        shutil.rmtree(
            get_absolute_file_path(f"pictures/{user.id}"),
            ignore_errors=True,
        )
        return {"status": "no content"}, 204
    except (
        exc.IntegrityError,
        exc.OperationalError,
        ValueError,
        OSError,
    ) as e:
        return handle_error_and_return_response(e, db=db)


@users_blueprint.route("/users/<user_name>/follow", methods=["POST"])
@require_auth(scopes=["follow:write"])
def follow_user(auth_user: User, user_name: str) -> Union[Dict, HttpResponse]:
    """
    Send a follow request to a user.

    **Scope**: ``follow:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/users/john_doe/follow HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "status": "success",
        "message": "Follow request to user 'john_doe' is sent.",
      }


    :param string user_name: user name

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user does not exist``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    successful_response_dict = {
        "status": "success",
        "message": f"Follow request to user '{user_name}' is sent.",
    }

    target_user = User.query.filter(
        func.lower(User.username) == func.lower(user_name),
    ).first()
    if not target_user:
        appLog.error(
            f"Error when following a user: user {user_name} not found"
        )
        return UserNotFoundErrorResponse()

    if auth_user.is_blocked_by(target_user):
        return InvalidPayloadErrorResponse("you can not follow this user")

    try:
        auth_user.send_follow_request_to(target_user)
    except FollowRequestAlreadyRejectedError:
        return ForbiddenErrorResponse()
    return successful_response_dict


@users_blueprint.route("/users/<user_name>/unfollow", methods=["POST"])
@require_auth(scopes=["follow:write"])
def unfollow_user(
    auth_user: User, user_name: str
) -> Union[Dict, HttpResponse]:
    """
    Unfollow a user.

    **Scope**: ``follow:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/users/john_doe/unfollow HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "status": "success",
        "message": "Undo for a follow request to user 'john_doe' is sent.",
      }


    :param string user_name: user name

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user does not exist``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    successful_response_dict = {
        "status": "success",
        "message": f"Undo for a follow request to user '{user_name}' is sent.",
    }

    target_user = User.query.filter(
        func.lower(User.username) == func.lower(user_name),
    ).first()
    if not target_user:
        appLog.error(
            f"Error when following a user: user {user_name} not found"
        )
        return UserNotFoundErrorResponse()

    try:
        auth_user.unfollows(target_user)
    except NotExistingFollowRequestError:
        return NotFoundErrorResponse(message="relationship does not exist")
    return successful_response_dict


def get_user_relationships(
    auth_user: User, user_name: str, relation: str
) -> Union[Dict, HttpResponse]:
    params = request.args.copy()
    try:
        page = int(params.get("page", 1))
    except ValueError:
        page = 1

    user = User.query.filter(
        func.lower(User.username) == func.lower(user_name),
    ).first()
    if not user:
        return UserNotFoundErrorResponse()

    relations_object = (
        user.followers if relation == "followers" else user.following
    )

    paginated_relations = relations_object.order_by(
        FollowRequest.updated_at.desc()
    ).paginate(page=page, per_page=USERS_PER_PAGE, error_out=False)

    return {
        "status": "success",
        "data": {
            relation: [
                user.serialize(current_user=auth_user)
                for user in paginated_relations.items
            ]
        },
        "pagination": {
            "has_next": paginated_relations.has_next,
            "has_prev": paginated_relations.has_prev,
            "page": paginated_relations.page,
            "pages": paginated_relations.pages,
            "total": paginated_relations.total,
        },
    }


@users_blueprint.route("/users/<user_name>/followers", methods=["GET"])
@require_auth(scopes=["follow:read"])
def get_followers(
    auth_user: User, user_name: str
) -> Union[Dict, HttpResponse]:
    """
    Get user followers.
    If the authenticated user has admin rights, it returns following users with
    additional field 'email'

    **Scope**: ``follow:read``

    **Example request**:

    - without parameters

    .. sourcecode:: http

      GET /api/users/sam/followers HTTP/1.1
      Content-Type: application/json

    - with page parameter

    .. sourcecode:: http

      GET /api/users/sam/followers?page=1 HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "followers": [
            {
              "admin": false,
              "bio": null,
              "birth_date": null,
              "created_at": "Thu, 02 Dec 2021 17:50:48 GMT",
              "first_name": null,
              "followers": 1,
              "following": 1,
              "follows": "true",
              "is_followed_by": "false",
              "last_name": null,
              "location": null,
              "map_visibility": "followers_only",
              "nb_sports": 0,
              "nb_workouts": 0,
              "picture": false,
              "records": [],
              "sports_list": [],
              "total_distance": 0.0,
              "total_duration": "0:00:00",
              "username": "JohnDoe",
              "workouts_visibility": "followers_only"
            }
          ]
        },
        "pagination": {
          "has_next": false,
          "has_prev": false,
          "page": 1,
          "pages": 1,
          "total": 1
        },
        "status": "success"
      }

    :param string user_name: user name

    :query integer page: page if using pagination (default: 1)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user does not exist``

    """
    return get_user_relationships(auth_user, user_name, "followers")


@users_blueprint.route("/users/<user_name>/following", methods=["GET"])
@require_auth(scopes=["follow:read"])
def get_following(
    auth_user: User, user_name: str
) -> Union[Dict, HttpResponse]:
    """
    Get user following.
    If the authenticate user has admin rights, it returns following users with
    additional field 'email'

    **Scope**: ``follow:read``

    **Example request**:

    - without parameters

    .. sourcecode:: http

      GET /api/users/sam/following HTTP/1.1
      Content-Type: application/json

    - with page parameter

    .. sourcecode:: http

      GET /api/users/sam/following?page=1 HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "following": [
            {
              "admin": false,
              "bio": null,
              "birth_date": null,
              "created_at": "Thu, 02 Dec 2021 17:50:48 GMT",
              "first_name": null,
              "followers": 1,
              "following": 1,
              "follows": "false",
              "is_followed_by": "true",
              "last_name": null,
              "location": null,
              "map_visibility": "followers_only",
              "nb_sports": 0,
              "nb_workouts": 0,
              "picture": false,
              "records": [],
              "sports_list": [],
              "total_distance": 0.0,
              "total_duration": "0:00:00",
              "username": "JohnDoe",
              "workouts_visibility": "followers_only"
            }
          ]
        },
        "pagination": {
          "has_next": false,
          "has_prev": false,
          "page": 1,
          "pages": 1,
          "total": 1
        },
        "status": "success"
      }

    :param string user_name: user name

    :query integer page: page if using pagination (default: 1)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user does not exist``

    """
    return get_user_relationships(auth_user, user_name, "following")


@users_blueprint.route("/users/<user_name>/block", methods=["POST"])
@require_auth(scopes=["users:write"])
def block_user(auth_user: User, user_name: str) -> Union[Dict, HttpResponse]:
    """
    Block a user

    **Scope**: ``users:write``

    **Example request**:

    .. sourcecode:: http

      GET /api/users/sam/block HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "status": "success"
      }

    :param string user_name: user name

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 400:
        - ``invalid payload``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user not found``
    """
    target_user = User.query.filter(
        func.lower(User.username) == func.lower(user_name),
    ).first()
    if not target_user:
        appLog.error(f"Error: user {user_name} not found")
        return UserNotFoundErrorResponse()

    try:
        auth_user.blocks_user(target_user)
        # delete follow request is exists (approved or pending)
        FollowRequest.query.filter_by(
            follower_user_id=target_user.id, followed_user_id=auth_user.id
        ).delete()
        db.session.commit()

    except BlockUserException:
        return InvalidPayloadErrorResponse()

    return {"status": "success"}


@users_blueprint.route("/users/<user_name>/unblock", methods=["POST"])
@require_auth(scopes=["users:write"])
def unblock_user(auth_user: User, user_name: str) -> Union[Dict, HttpResponse]:
    """
    Unblock a user

    **Scope**: ``users:write``

    **Example request**:

    .. sourcecode:: http

      GET /api/users/sam/unblock HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "status": "success"
      }

    :param string user_name: user name

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user not found``

    """
    target_user = User.query.filter(
        func.lower(User.username) == func.lower(user_name),
    ).first()
    if not target_user:
        appLog.error(f"Error: user {user_name} not found")
        return UserNotFoundErrorResponse()

    auth_user.unblocks_user(target_user)

    return {"status": "success"}


@users_blueprint.route("/users/<user_name>/sanctions", methods=["GET"])
@require_auth(scopes=["users:read"], allow_suspended_user=True)
def get_user_sanctions(
    auth_user: User, user_name: str
) -> Union[Dict, HttpResponse]:
    """
    Get user sanctions.

    It returns sanctions only if:
    - user name is authenticated user username
    - user has moderation rights.

    Suspended user can access this endpoint.

    **Scope**: ``users:read``

    **Example requests**:

    - without parameters:

    .. sourcecode:: http

      GET /api/users/Sam/sanctions HTTP/1.1

    - with parameters:

    .. sourcecode:: http

      GET /api/users/Sam/sanctions?page=2 HTTP/1.1

    **Example responses**:

    - if sanctions exist (response with moderation rights)

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "data": {
            "sanctions": [
              {
                "action_type": "workout_suspension",
                "appeal": {
                  "approved": null,
                  "created_at": "Wed, 04 Dec 2024 11:00:04 GMT",
                  "id": "2ULe2hWhSnYCS2VHbsikB9",
                  "moderator": null,
                  "reason": null,
                  "text": "<APPEAL TEXT>",
                  "updated_at": null,
                  "user": {
                    "blocked": false,
                    "created_at": "Wed, 04 Dec 2024 09:07:06 GMT",
                    "email": "sam@example.com",
                    "followers": 0,
                    "following": 0,
                    "follows": false,
                    "is_active": true,
                    "is_followed_by": false,
                    "nb_workouts": 1,
                    "picture": false,
                    "role": "user",
                    "suspended_at": null,
                    "username": "Sam"
                  }
                },
                "created_at": "Wed, 04 Dec 2024 10:59:45 GMT",
                "id": "6dxczvMrhkAR72shUz9Pwd",
                "moderator": {
                  "blocked": false,
                  "created_at": "Wed, 01 Mar 2023 12:31:17 GMT",
                  "email": "admin@example.com",
                  "followers": 0,
                  "following": 0,
                  "follows": "false",
                  "is_active": true,
                  "is_followed_by": "false",
                  "nb_workouts": 0,
                  "picture": true,
                  "role": "admin",
                  "suspended_at": null,
                  "username": "admin"
                },
                "reason": "<SUSPENSION REASON>",
                "report_id": 2,
                "user": {
                  "blocked": false,
                  "created_at": "Sun, 01 Dec 2024 17:27:49 GMT",
                  "email": "sam@example.com",
                  "followers": 0,
                  "following": 0,
                  "follows": "false",
                  "is_active": true,
                  "is_followed_by": "false",
                  "nb_workouts": 1,
                  "picture": false,
                  "role": "user",
                  "suspended_at": null,
                  "username": "Sam"
                }
              }
            ]
          },
          "pagination": {
            "has_next": false,
            "has_prev": false,
            "page": 1,
            "pages": 1,
            "total": 1
          },
          "status": "success"
        }

    - if sanctions exist (response for authenticated user)

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "data": {
            "sanctions": [
              {
                "action_type": "workout_suspension",
                "appeal": {
                  "approved": null,
                  "created_at": "Wed, 04 Dec 2024 16:50:55 GMT",
                  "id": "kcj6hdGQqPKaaKQmfQj8Jv",
                  "reason": null,
                  "text": "<APPEAL TEXT>",
                  "updated_at": null
                },
                "created_at": "Wed, 04 Dec 2024 16:50:44 GMT",
                "id": "6nvxvAyoh9Zkr8RMXhu54T",
                "reason": "<SUSPENSION REASON>"
              }
            ]
          },
          "pagination": {
            "has_next": false,
            "has_prev": false,
            "page": 1,
            "pages": 1,
            "total": 1
          },
          "status": "success"
        }

    - no sanctions

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "data": {
            "sanctions": []
          },
          "pagination": {
            "has_next": false,
            "has_prev": false,
            "page": 1,
            "pages": 0,
            "total": 0
          },
          "status": "success"
        }

    :param string user_name: user name

    :query integer page: page if using pagination (default: 1)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
    :statuscode 404:
        - ``user not found``

    """
    user = User.query.filter(
        func.lower(User.username) == func.lower(user_name),
    ).first()
    if not user:
        appLog.error(f"Error: user {user_name} not found")
        return UserNotFoundErrorResponse()

    if user.id != auth_user.id and not auth_user.has_moderator_rights:
        return ForbiddenErrorResponse()

    params = request.args.copy()
    page = int(params.get("page", 1))

    paginated_sanctions = (
        ReportAction.query.filter(
            ReportAction.user_id == user.id,
            ReportAction.action_type.not_in(
                [
                    "comment_unsuspension",
                    "user_unsuspension",
                    "user_warning_lifting",
                    "workout_unsuspension",
                ]
            ),
        )
        .order_by(ReportAction.created_at.desc())
        .paginate(page=page, per_page=ACTIONS_PER_PAGE, error_out=False)
    )

    return {
        "status": "success",
        "data": {
            "sanctions": [
                sanctions.serialize(current_user=auth_user, full=False)
                for sanctions in paginated_sanctions.items
            ]
        },
        "pagination": {
            "has_next": paginated_sanctions.has_next,
            "has_prev": paginated_sanctions.has_prev,
            "page": paginated_sanctions.page,
            "pages": paginated_sanctions.pages,
            "total": paginated_sanctions.total,
        },
    }


@users_blueprint.route("/users/<user_name>/workouts", methods=["GET"])
@require_auth(scopes=["workouts:read"], optional_auth_user=True)
def get_user_latest_workouts(
    auth_user: User, user_name: str
) -> Union[Dict, HttpResponse]:
    """
    Get user last 5 visible workouts.

    **Scope**: ``workouts:read``

    **Example request**:

      GET /api/users/Sam/workouts HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "data": {
            "workouts": [
              {
                "ascent": null,
                "ave_speed": 10.0,
                "bounds": [],
                "creation_date": "Sun, 14 Jul 2019 13:51:01 GMT",
                "descent": null,
                "description": null,
                "distance": 10.0,
                "duration": "0:17:04",
                "equipments": [],
                "id": "kjxavSTUrJvoAh2wvCeGEF",
                "liked": false,
                "likes_count": 0,
                "map": null,
                "map_visibility": "private",
                "max_alt": null,
                "max_speed": 10.0,
                "min_alt": null,
                "modification_date": null,
                "moving": "0:17:04",
                "next_workout": 3,
                "notes": null,
                "pauses": null,
                "previous_workout": null,
                "records": [
                  {
                    "id": 4,
                    "record_type": "MS",
                    "sport_id": 1,
                    "user": "admin",
                    "value": 10.0,
                    "workout_date": "Mon, 01 Jan 2018 00:00:00 GMT",
                    "workout_id": "kjxavSTUrJvoAh2wvCeGEF"
                  },
                  {
                    "id": 13,
                    "record_type": "HA",
                    "sport_id": 1,
                    "user": "Sam",
                    "value": 43.97,
                    "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                    "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                  },
                  {
                    "id": 3,
                    "record_type": "LD",
                    "sport_id": 1,
                    "user": "admin",
                    "value": "0:17:04",
                    "workout_date": "Mon, 01 Jan 2018 00:00:00 GMT",
                    "workout_id": "kjxavSTUrJvoAh2wvCeGEF"
                  },
                  {
                    "id": 2,
                    "record_type": "FD",
                    "sport_id": 1,
                    "user": "admin",
                    "value": 10.0,
                    "workout_date": "Mon, 01 Jan 2018 00:00:00 GMT",
                    "workout_id": "kjxavSTUrJvoAh2wvCeGEF"
                  },
                  {
                    "id": 1,
                    "record_type": "AS",
                    "sport_id": 1,
                    "user": "admin",
                    "value": 10.0,
                    "workout_date": "Mon, 01 Jan 2018 00:00:00 GMT",
                    "workout_id": "kjxavSTUrJvoAh2wvCeGEF"
                  }
                ],
                "segments": [],
                "sport_id": 1,
                "suspended": false,
                "suspended_at": null,
                "title": null,
                "user": {
                  "created_at": "Sun, 31 Dec 2017 09:00:00 GMT",
                  "followers": 0,
                  "following": 0,
                  "nb_workouts": 1,
                  "picture": false,
                  "role": "user",
                  "suspended_at": null,
                  "username": "Sam"
                },
                "weather_end": null,
                "weather_start": null,
                "with_gpx": false,
                "workout_date": "Mon, 01 Jan 2018 00:00:00 GMT",
                "workout_visibility": "private"
              }
            ]
          },
          "status": "success"
        }

    :param string user_name: user name


    :reqheader Authorization: OAuth 2.0 Bearer Token if user is authenticated

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
    :statuscode 404:
        - ``user not found``
    """
    user = User.query.filter(
        func.lower(User.username) == func.lower(user_name),
    ).first()
    if not user:
        appLog.error(f"Error: user {user_name} not found")
        return UserNotFoundErrorResponse()
    if user.suspended_at:
        return {"status": "success", "data": {"workouts": []}}

    workouts_query = Workout.query.filter(
        Workout.suspended_at == None,  # noqa
        Workout.user_id == user.id,
    )
    if not auth_user or (
        auth_user.id != user.id
        and auth_user.id not in user.get_followers_user_ids()
    ):
        workouts_query = workouts_query.filter(
            Workout.workout_visibility == VisibilityLevel.PUBLIC
        )
    elif auth_user.id in user.get_followers_user_ids():
        workouts_query = workouts_query.filter(
            Workout.workout_visibility.in_(
                [VisibilityLevel.PUBLIC, VisibilityLevel.FOLLOWERS]
            )
        )
    workouts = (
        workouts_query.order_by(Workout.workout_date.desc())
        .limit(WORKOUTS_PER_PAGE)
        .all()
    )
    return {
        "status": "success",
        "data": {
            "workouts": [
                workout.serialize(user=auth_user) for workout in workouts
            ]
        },
    }



from typing import Optional


class GenericException(Exception):
    def __init__(
        self, status: str, message: str, e: Optional[Exception] = None
    ) -> None:
        super().__init__(message)
        self.status = status
        self.message = message
        self.e = e



"""Add equipments and equipment types for workouts

Revision ID: 171dcd7e5f2b
Revises: 24eb097614e4
Create Date: 2023-03-20 22:50:47.672811

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql


# revision identifiers, used by Alembic.
revision = '171dcd7e5f2b'
down_revision = '4d51a4ca8001'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'equipment_types',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('label', sa.String(length=50), nullable=False),
        sa.Column('is_active', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
    )
    op.execute(
        """
        INSERT INTO equipment_types (label, is_active)
        VALUES 
        ('Shoes', True), ('Bike', True), 
        ('Bike Trainer', True), ('Kayak_Boat', True),
        ('Skis', True), ('Snowshoes', True) 
        """
    )

    op.create_table(
        'equipments',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('label', sa.String(length=50), nullable=False),
        sa.Column('description', sa.String(length=200), nullable=True),
        sa.Column('equipment_type_id', sa.Integer(), nullable=True),
        sa.Column('creation_date', sa.DateTime(), nullable=True),
        sa.Column('is_active', sa.Boolean(), nullable=False),
        sa.Column(
            'total_distance',
            sa.Numeric(precision=10, scale=3),
            server_default=sa.text('0.0'),
            nullable=True,
        ),
        sa.Column(
            'total_duration',
            sa.Interval(),
            server_default=sa.text("'00:00:00'"),
            nullable=False,
        ),
        sa.Column(
            'total_moving',
            sa.Interval(),
            server_default=sa.text("'00:00:00'"),
            nullable=False,
        ),
        sa.Column(
            'total_workouts',
            sa.Integer(),
            server_default=sa.text('0'),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ['equipment_type_id'],
            ['equipment_types.id'],
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['users.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'label',
            'user_id',
            name='equipment_user_label_unique',
        ),
        sa.UniqueConstraint(
            'uuid',
            name='equipments_uuid_key',
        ),
    )
    with op.batch_alter_table('equipments', schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f('ix_equipments_user_id'), ['user_id'], unique=False
        )

    op.create_table(
        'workout_equipments',
        sa.Column('workout_id', sa.Integer(), nullable=False),
        sa.Column('equipment_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['workout_id'], ['workouts.id'], ondelete='CASCADE'
        ),
        sa.ForeignKeyConstraint(
            ['equipment_id'], ['equipments.id'], ondelete='CASCADE'
        ),
        sa.PrimaryKeyConstraint('workout_id', 'equipment_id'),
    )

    op.create_table(
        'users_sports_preferences_equipments',
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('sport_id', sa.Integer(), nullable=False),
        sa.Column('equipment_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['equipment_id'], ['equipments.id'], ondelete='CASCADE'
        ),
        sa.ForeignKeyConstraint(
            ['sport_id'], ['sports.id'], ondelete='CASCADE'
        ),
        sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('user_id', 'sport_id', 'equipment_id'),
    )


def downgrade():
    op.drop_table('users_sports_preferences_equipments')

    op.drop_table('workout_equipments')
    with op.batch_alter_table('equipments', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_equipments_user_id'))

    op.drop_table('equipments')
    op.drop_table('equipment_types')



from fittrackee.exceptions import GenericException


class InvalidDurationException(Exception):
    def __init__(self) -> None:
        super().__init__("invalid duration")


class InvalidGPXException(GenericException):
    pass


class WorkoutException(GenericException):
    pass


class WorkoutGPXException(GenericException):
    pass


class WorkoutForbiddenException(GenericException):
    def __init__(self) -> None:
        super().__init__("error", "you do not have permissions")



import os
from datetime import datetime, timedelta, timezone
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
from uuid import UUID, uuid4

from sqlalchemy.dialects import postgresql
from sqlalchemy.engine.base import Connection
from sqlalchemy.event import listens_for
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.orm.session import Session, object_session
from sqlalchemy.sql.expression import nulls_last
from sqlalchemy.types import JSON, Enum

from fittrackee import BaseModel, appLog, db
from fittrackee.database import TZDateTime
from fittrackee.dates import aware_utc_now
from fittrackee.equipments.models import WorkoutEquipment
from fittrackee.files import get_absolute_file_path
from fittrackee.utils import encode_uuid
from fittrackee.visibility_levels import (
    VisibilityLevel,
    can_view,
    get_calculated_visibility,
)

from .exceptions import WorkoutForbiddenException
from .utils.convert import convert_in_duration, convert_value_to_integer

if TYPE_CHECKING:
    from sqlalchemy.orm.attributes import AttributeEvent

    from fittrackee.comments.models import Comment
    from fittrackee.equipments.models import Equipment
    from fittrackee.reports.models import Report, ReportAction
    from fittrackee.users.models import User


EMPTY_MINIMAL_WORKOUT_VALUES: Dict = {
    "title": "",
    "moving": None,
    "distance": None,
    "duration": None,
    "min_alt": None,
    "max_alt": None,
    "descent": None,
    "ascent": None,
    "map_visibility": None,
}
EMPTY_WORKOUT_VALUES: Dict = {
    "creation_date": None,
    "modification_date": None,
    "pauses": None,
    "equipments": [],
    "records": [],
    "segments": [],
    "weather_start": None,
    "weather_end": None,
    "notes": "",
    "likes_count": 0,
    "liked": False,
}

record_types = [
    "AS",  # 'Best Average Speed'
    "FD",  # 'Farthest Distance'
    "HA",  # 'Highest Ascent'
    "LD",  # 'Longest Duration'
    "MS",  # 'Max speed'
]
DESCRIPTION_MAX_CHARACTERS = 10000
NOTES_MAX_CHARACTERS = 500
TITLE_MAX_CHARACTERS = 255


def update_records(
    user_id: int, sport_id: int, connection: Connection, session: Session
) -> None:
    record_table = Record.__table__  # type: ignore
    new_records = Workout.get_user_workout_records(user_id, sport_id)
    for record_type, record_data in new_records.items():
        if record_data["record_value"]:
            record = Record.query.filter_by(
                user_id=user_id, sport_id=sport_id, record_type=record_type
            ).first()
            if record:
                value = convert_value_to_integer(
                    record_type, record_data["record_value"]
                )
                connection.execute(
                    record_table.update()
                    .where(record_table.c.id == record.id)
                    .values(
                        value=value,
                        workout_id=record_data["workout"].id,
                        workout_uuid=record_data["workout"].uuid,
                        workout_date=record_data["workout"].workout_date,
                    )
                )
            else:
                new_record = Record(
                    workout=record_data["workout"], record_type=record_type
                )
                new_record.value = record_data["record_value"]  # type: ignore
                session.add(new_record)
        else:
            connection.execute(
                record_table.delete()
                .where(record_table.c.user_id == user_id)
                .where(record_table.c.sport_id == sport_id)
                .where(record_table.c.record_type == record_type)
            )


def format_value(
    value: Union[Decimal, timedelta], attribute: str
) -> Union[float, timedelta]:
    return float(value) if attribute == "distance" else value  # type: ignore


def update_equipments(workout: "Workout", connection: Connection) -> None:
    from fittrackee.equipments.models import Equipment

    instance_state = db.inspect(workout)
    workout_values = {}

    for attribute in ["distance", "duration", "moving"]:
        state_history = instance_state.attrs[attribute].load_history()
        if len(state_history.added) > 0 and len(state_history.deleted) > 0:
            workout_values[attribute] = {
                "new": format_value(state_history.added[0], attribute),
                "old": format_value(state_history.deleted[0], attribute),
            }
    if not workout_values:
        return

    equipment_table = Equipment.__table__  # type: ignore
    for equipment in workout.equipments:
        equipment_values = {}
        for attribute, value in workout_values.items():
            column = getattr(equipment, f"total_{attribute}")
            equipment_values[f"total_{attribute}"] = (
                format_value(column, attribute)  # type: ignore
                - value["old"]
                + value["new"]
            )

        connection.execute(
            equipment_table.update()
            .where(equipment_table.c.id == equipment.id)
            .values(**equipment_values)
        )


class Sport(BaseModel):
    __tablename__ = "sports"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    label: Mapped[str] = mapped_column(
        db.String(50), unique=True, nullable=False
    )
    is_active: Mapped[bool] = mapped_column(default=True, nullable=False)
    stopped_speed_threshold: Mapped[float] = mapped_column(
        default=1.0, nullable=False
    )

    workouts: Mapped[List["Workout"]] = relationship(
        "Workout", lazy=True, back_populates="sport"
    )
    records: Mapped[List["Record"]] = relationship(
        "Record", lazy=True, back_populates="sport"
    )

    def __repr__(self) -> str:
        return f"<Sport {self.label!r}>"

    def __init__(self, label: str) -> None:
        self.label = label

    @property
    def has_workouts(self) -> bool:
        return (
            db.session.query(Workout.id)
            .filter_by(sport_id=self.id)
            .limit(1)
            .count()
            > 0
        )

    def serialize(
        self,
        *,
        check_workouts: bool = False,
        sport_preferences: Optional[Dict] = None,
    ) -> Dict:
        serialized_sport = {
            "id": self.id,
            "label": self.label,
            "is_active": self.is_active,
            "is_active_for_user": (
                self.is_active
                if sport_preferences is None
                else (sport_preferences["is_active"] and self.is_active)
            ),
            "color": (
                None
                if sport_preferences is None
                else sport_preferences["color"]
            ),
            "stopped_speed_threshold": (
                self.stopped_speed_threshold
                if sport_preferences is None
                else sport_preferences["stopped_speed_threshold"]
            ),
        }
        if check_workouts:
            serialized_sport["has_workouts"] = self.has_workouts

        serialized_sport["default_equipments"] = (
            []
            if sport_preferences is None
            else sport_preferences["default_equipments"]
        )

        return serialized_sport


class Workout(BaseModel):
    __tablename__ = "workouts"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    uuid: Mapped[UUID] = mapped_column(
        postgresql.UUID(as_uuid=True),
        default=uuid4,
        unique=True,
        nullable=False,
    )
    user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id"), index=True, nullable=False
    )
    sport_id: Mapped[int] = mapped_column(
        db.ForeignKey("sports.id"), index=True, nullable=False
    )
    title: Mapped[Optional[str]] = mapped_column(
        db.String(TITLE_MAX_CHARACTERS), nullable=True
    )
    gpx: Mapped[Optional[str]] = mapped_column(db.String(255), nullable=True)
    creation_date: Mapped[datetime] = mapped_column(
        TZDateTime, default=aware_utc_now
    )
    modification_date: Mapped[Optional[datetime]] = mapped_column(
        TZDateTime, onupdate=aware_utc_now, nullable=True
    )
    workout_date: Mapped[datetime] = mapped_column(
        TZDateTime, index=True, nullable=False
    )
    duration: Mapped[timedelta] = mapped_column(nullable=False)
    pauses: Mapped[Optional[timedelta]] = mapped_column(nullable=True)
    moving: Mapped[Optional[timedelta]] = mapped_column(nullable=True)
    distance: Mapped[Optional[float]] = mapped_column(
        db.Numeric(6, 3), nullable=True
    )  # kilometers
    min_alt: Mapped[Optional[float]] = mapped_column(
        db.Numeric(6, 2), nullable=True
    )  # meters
    max_alt: Mapped[Optional[float]] = mapped_column(
        db.Numeric(6, 2), nullable=True
    )  # meters
    descent: Mapped[Optional[float]] = mapped_column(
        db.Numeric(8, 3), nullable=True
    )  # meters
    ascent: Mapped[Optional[float]] = mapped_column(
        db.Numeric(8, 3), nullable=True
    )  # meters
    max_speed: Mapped[Optional[float]] = mapped_column(
        db.Numeric(6, 2), nullable=True
    )  # km/h
    ave_speed: Mapped[Optional[float]] = mapped_column(
        db.Numeric(6, 2), nullable=True
    )  # km/h
    bounds: Mapped[Optional[List[float]]] = mapped_column(
        postgresql.ARRAY(db.Float), nullable=True
    )
    map: Mapped[Optional[str]] = mapped_column(db.String(255), nullable=True)
    map_id: Mapped[Optional[str]] = mapped_column(
        db.String(50), index=True, nullable=True
    )
    weather_start: Mapped[Optional[Dict]] = mapped_column(JSON, nullable=True)
    weather_end: Mapped[Optional[Dict]] = mapped_column(JSON, nullable=True)
    notes: Mapped[Optional[str]] = mapped_column(
        db.String(NOTES_MAX_CHARACTERS), nullable=True
    )
    description: Mapped[Optional[str]] = mapped_column(
        db.String(DESCRIPTION_MAX_CHARACTERS), nullable=True
    )
    workout_visibility: Mapped[VisibilityLevel] = mapped_column(
        Enum(VisibilityLevel, name="visibility_levels"),
        server_default="PRIVATE",
        nullable=False,
    )
    map_visibility: Mapped[VisibilityLevel] = mapped_column(
        Enum(VisibilityLevel, name="visibility_levels"),
        server_default="PRIVATE",
        nullable=False,
    )
    suspended_at: Mapped[Optional[datetime]] = mapped_column(
        TZDateTime, nullable=True
    )
    analysis_visibility: Mapped[VisibilityLevel] = mapped_column(
        Enum(VisibilityLevel, name="visibility_levels"),
        server_default="PRIVATE",
        nullable=False,
    )

    user: Mapped["User"] = relationship(
        "User", lazy="select", single_parent=True
    )
    sport: Mapped["Sport"] = relationship(
        "Sport", lazy="joined", single_parent=True
    )
    segments: Mapped[List["WorkoutSegment"]] = relationship(
        "WorkoutSegment",
        lazy=True,
        cascade="all, delete",
        back_populates="workout",
    )
    records: Mapped[List["Record"]] = relationship(
        "Record",
        lazy=True,
        cascade="all, delete",
        back_populates="workout",
        order_by="Record.record_type.asc()",
    )
    equipments: Mapped[List["Equipment"]] = relationship(
        "Equipment", secondary=WorkoutEquipment, back_populates="workouts"
    )
    comments: Mapped[List["Comment"]] = relationship(
        "Comment",
        lazy=True,
        back_populates="workout",
    )
    likes = relationship(
        "User",
        secondary="workout_likes",
        primaryjoin="Workout.id == WorkoutLike.workout_id",
        secondaryjoin="WorkoutLike.user_id == User.id",
        lazy="dynamic",
        viewonly=True,
    )

    def __str__(self) -> str:
        return f"<Workout '{self.sport.label}' - {self.workout_date}>"

    def __init__(
        self,
        user_id: int,
        sport_id: int,
        workout_date: datetime,
        distance: float,
        duration: timedelta,
    ) -> None:
        self.user_id = user_id
        self.sport_id = sport_id
        self.workout_date = workout_date
        self.distance = distance
        self.duration = duration

    @property
    def short_id(self) -> str:
        return encode_uuid(self.uuid)

    @property
    def calculated_analysis_visibility(self) -> VisibilityLevel:
        return get_calculated_visibility(
            visibility=self.analysis_visibility,
            parent_visibility=self.workout_visibility,
        )

    @property
    def calculated_map_visibility(self) -> VisibilityLevel:
        return get_calculated_visibility(
            visibility=self.map_visibility,
            parent_visibility=self.analysis_visibility,
        )

    def liked_by(self, user: "User") -> bool:
        return user in self.likes.all()

    @property
    def suspension_action(self) -> Optional["ReportAction"]:
        if self.suspended_at is None:
            return None

        from fittrackee.reports.models import ReportAction

        return (
            ReportAction.query.filter(
                ReportAction.workout_id == self.id,
                ReportAction.action_type == "workout_suspension",
            )
            .order_by(ReportAction.created_at.desc())
            .first()
        )

    @property
    def reports(self) -> List["Report"]:
        from fittrackee.reports.models import Report

        return Report.query.filter_by(reported_workout_id=self.id).all()

    def get_workout_data(
        self,
        user: Optional["User"],
        *,
        can_see_analysis_data: Optional[bool] = None,
        can_see_map_data: Optional[bool] = None,
        for_report: bool = False,
        additional_data: bool = False,
        light: bool = True,
        with_equipments: bool = False,  # for workouts list
    ) -> Dict:
        """
        Used by Workout serializer and data export

        - can_see_analysis_data: if user can see charts
        - can_see_map_data: if user can see map
        - for_report: privacy levels are overridden on report
        - additional_data is False when:
          - workout is not suspended
          - user is workout owner
            or workout is displayed in report and current user has moderation
            rights
        - light: when true, only workouts data needed for workout lists
          and timeline.
          If 'light' is False, 'with_equipments' is ignored.
        - with_equipments: only used when 'light' is True. Needed for
          3rd-party apps updating workouts equipments
        """
        for_report = (
            for_report and user is not None and user.has_moderator_rights
        )
        if can_see_analysis_data is None:
            can_see_analysis_data = can_view(
                self,
                "calculated_analysis_visibility",
                user=user,
                for_report=for_report,
            )
        if can_see_map_data is None:
            can_see_map_data = can_view(
                self,
                "calculated_map_visibility",
                user=user,
                for_report=for_report,
            )

        workout_data = {
            "id": self.short_id,  # WARNING: client use uuid as id
            "sport_id": self.sport_id,
            "workout_date": self.workout_date,
            "workout_visibility": self.workout_visibility.value,
        }

        if not additional_data:
            return {
                **workout_data,
                **EMPTY_MINIMAL_WORKOUT_VALUES,
                **EMPTY_WORKOUT_VALUES,
            }

        workout_data = {
            **workout_data,
            **EMPTY_WORKOUT_VALUES,
            "title": self.title,
            "moving": None if self.moving is None else str(self.moving),
            "distance": (
                None if self.distance is None else float(self.distance)
            ),
            "duration": (
                None if self.duration is None else str(self.duration)
            ),
            "ave_speed": (
                None if self.ave_speed is None else float(self.ave_speed)
            ),
            "max_speed": (
                None if self.max_speed is None else float(self.max_speed)
            ),
            "min_alt": (
                float(self.min_alt)
                if self.min_alt is not None and can_see_analysis_data
                else None
            ),
            "max_alt": (
                float(self.max_alt)
                if self.max_alt is not None and can_see_analysis_data
                else None
            ),
            "descent": None if self.descent is None else float(self.descent),
            "ascent": None if self.ascent is None else float(self.ascent),
            "analysis_visibility": (
                self.calculated_analysis_visibility.value
                if can_see_analysis_data
                else VisibilityLevel.PRIVATE
            ),
            "map_visibility": (
                self.calculated_map_visibility.value
                if can_see_map_data
                else VisibilityLevel.PRIVATE
            ),
        }

        if not light or with_equipments:
            workout_data["equipments"] = [
                equipment.serialize(current_user=user)
                for equipment in self.equipments
                if can_view(equipment, "visibility", user)
            ]

        if light:
            return workout_data

        return {
            **workout_data,
            "creation_date": self.creation_date,
            "modification_date": self.modification_date,
            "pauses": str(self.pauses) if self.pauses else None,
            "records": (
                []
                if for_report
                else [record.serialize() for record in self.records]
            ),
            "segments": (
                [segment.serialize() for segment in self.segments]
                if can_see_analysis_data
                else []
            ),
            "weather_start": self.weather_start,
            "weather_end": self.weather_end,
            "notes": (
                self.notes if user and user.id == self.user_id else None
            ),
            "description": self.description,
            "likes_count": self.likes.count(),
            "liked": self.liked_by(user) if user else False,
        }

    def serialize(
        self,
        *,
        user: Optional["User"] = None,
        params: Optional[Dict] = None,
        for_report: bool = False,
        light: bool = True,  # for workouts list and timeline
        with_equipments: bool = False,  # for workouts list
    ) -> Dict:
        """
        If 'light' is False, 'with_equipments' is ignored.
        """

        for_report = (
            for_report and user is not None and user.has_moderator_rights
        )
        if not can_view(
            self, "workout_visibility", user=user, for_report=for_report
        ):
            raise WorkoutForbiddenException()
        can_see_analysis_data = can_view(
            self,
            "calculated_analysis_visibility",
            user=user,
            for_report=for_report,
        )
        can_see_map_data = can_view(
            self, "calculated_map_visibility", user=user, for_report=for_report
        )
        is_owner = user is not None and user.id == self.user_id
        is_workout_suspended = self.suspended_at is not None
        additional_data = not is_workout_suspended or for_report or is_owner

        workout = self.get_workout_data(
            user,
            can_see_analysis_data=can_see_analysis_data,
            can_see_map_data=can_see_map_data,
            for_report=for_report,
            additional_data=additional_data,
            light=light,
            with_equipments=with_equipments,
        )

        workout["map"] = (
            self.map_id
            if self.map and can_see_map_data and additional_data
            else None
        )
        workout["with_gpx"] = (
            self.gpx is not None and can_see_map_data and additional_data
        )
        workout["with_analysis"] = (
            self.gpx is not None and can_see_analysis_data and additional_data
        )
        workout["suspended"] = is_workout_suspended
        workout["user"] = self.user.serialize()

        if is_owner or for_report:
            workout["suspended_at"] = self.suspended_at
            if self.suspension_action:
                workout["suspension"] = self.suspension_action.serialize(
                    current_user=user,  # type: ignore
                    full=False,
                )

        if light:
            workout["next_workout"] = None
            workout["previous_workout"] = None
            workout["bounds"] = []
            return workout

        if is_owner:
            if params and user:
                from .utils.workouts import get_datetime_from_request_args

                date_from, date_to = get_datetime_from_request_args(
                    params, user
                )
            else:
                date_from, date_to = (None, None)

            distance_from = params.get("distance_from") if params else None
            distance_to = params.get("distance_to") if params else None
            duration_from = params.get("duration_from") if params else None
            duration_to = params.get("duration_to") if params else None
            ave_speed_from = params.get("ave_speed_from") if params else None
            ave_speed_to = params.get("ave_speed_to") if params else None
            max_speed_from = params.get("max_speed_from") if params else None
            max_speed_to = params.get("max_speed_to") if params else None
            sport_id = params.get("sport_id") if params else None

            filters = [
                Workout.id != self.id,
                Workout.user_id == self.user_id,
            ]
            if sport_id:
                filters.append(Workout.sport_id == sport_id)
            if date_from:
                filters.append(Workout.workout_date >= date_from)
            if date_to:
                filters.append(Workout.workout_date <= date_to)
            if distance_from:
                filters.append(Workout.distance >= float(distance_from))
            if distance_to:
                filters.append(Workout.distance <= float(distance_to))
            if duration_from:
                filters.append(
                    Workout.duration >= convert_in_duration(duration_from)
                )
            if duration_to:
                filters.append(
                    Workout.duration <= convert_in_duration(duration_to)
                )
            if ave_speed_from:
                filters.append(Workout.ave_speed >= float(ave_speed_from))
            if ave_speed_to:
                filters.append(Workout.ave_speed <= float(ave_speed_to))
            if max_speed_from:
                filters.append(Workout.max_speed >= float(max_speed_from))
            if max_speed_to:
                filters.append(Workout.max_speed <= float(max_speed_to))

            previous_workout = (
                Workout.query.filter(
                    Workout.workout_date <= self.workout_date, *filters
                )
                .order_by(Workout.workout_date.desc())
                .first()
            )
            next_workout = (
                Workout.query.filter(
                    Workout.workout_date >= self.workout_date, *filters
                )
                .order_by(Workout.workout_date.asc())
                .first()
            )

        else:
            next_workout = None
            previous_workout = None

        workout["next_workout"] = (
            next_workout.short_id if next_workout else None
        )
        workout["previous_workout"] = (
            previous_workout.short_id if previous_workout else None
        )
        workout["bounds"] = (
            [float(bound) for bound in self.bounds]
            if self.bounds and can_see_map_data and additional_data
            else []
        )
        return workout

    @classmethod
    def get_user_workout_records(cls, user_id: int, sport_id: int) -> Dict:
        """
        Note:
        Values for ascent are null for workouts without gpx
        """
        record_types_columns = {
            "AS": "ave_speed",  # 'Average speed'
            "FD": "distance",  # 'Farthest Distance'
            "HA": "ascent",  # 'Highest Ascent'
            "LD": "moving",  # 'Longest Duration'
            "MS": "max_speed",  # 'Max speed'
        }
        records = {}
        for record_type, column in record_types_columns.items():
            column_sorted = getattr(Workout, column).desc()
            record_workout = (
                Workout.query.filter(
                    Workout.user_id == user_id, Workout.sport_id == sport_id
                )
                .order_by(nulls_last(column_sorted), Workout.workout_date)
                .first()
            )
            records[record_type] = dict(
                record_value=(
                    getattr(record_workout, column) if record_workout else None
                ),
                workout=record_workout,
            )
        return records


@listens_for(Workout, "after_insert")
def on_workout_insert(
    mapper: Mapper, connection: Connection, workout: Workout
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        update_records(workout.user_id, workout.sport_id, connection, session)


@listens_for(Workout, "after_update")
def on_workout_update(
    mapper: Mapper, connection: Connection, workout: Workout
) -> None:
    workout_object = object_session(workout)
    if workout_object and workout_object.is_modified(
        workout, include_collections=True
    ):

        @listens_for(db.Session, "after_flush", once=True)
        def receive_after_flush(session: Session, context: Any) -> None:
            if workout.equipments:
                update_equipments(workout, connection)
            sports_list = [workout.sport_id]
            records = Record.query.filter_by(workout_id=workout.id).all()
            for rec in records:
                if rec.sport_id not in sports_list:
                    sports_list.append(rec.sport_id)
            for sport_id in sports_list:
                update_records(workout.user_id, sport_id, connection, session)


@listens_for(Workout, "after_delete")
def on_workout_delete(
    mapper: Mapper, connection: Connection, old_workout: "Workout"
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        from fittrackee.users.models import Notification

        # Equipments must be removed before deleting workout
        # in order to recalculate equipments totals
        if old_workout.equipments:
            raise Exception("equipments exists, remove them first")

        if old_workout.map:
            try:
                os.remove(get_absolute_file_path(old_workout.map))
            except OSError:
                appLog.error("map file not found when deleting workout")
        if old_workout.gpx:
            try:
                os.remove(get_absolute_file_path(old_workout.gpx))
            except OSError:
                appLog.error("gpx file not found when deleting workout")

        Notification.query.filter(
            Notification.event_object_id == old_workout.id,
            Notification.to_user_id == old_workout.user_id,
            Notification.event_type.in_(
                [
                    "workout_comment",
                    "workout_like",
                ]
            ),
        ).delete()


@listens_for(Workout.equipments, "append")
def on_workout_equipments_append(
    target: Workout, value: "Equipment", initiator: "AttributeEvent"
) -> None:
    value.total_distance = float(value.total_distance) + (
        0.0 if target.distance is None else float(target.distance)
    )
    value.total_duration += target.duration
    if target.moving:
        value.total_moving += target.moving
    value.total_workouts += 1


@listens_for(Workout.equipments, "remove")
def on_workout_equipments_remove(
    target: Workout, value: "Equipment", initiator: "AttributeEvent"
) -> None:
    value.total_distance = float(value.total_distance) - (
        0.0 if target.distance is None else float(target.distance)
    )
    value.total_duration -= target.duration
    if target.moving:
        value.total_moving -= target.moving
    value.total_workouts -= 1


class WorkoutSegment(BaseModel):
    __tablename__ = "workout_segments"
    workout_id: Mapped[int] = mapped_column(
        db.ForeignKey("workouts.id"), primary_key=True
    )
    workout_uuid: Mapped[UUID] = mapped_column(
        postgresql.UUID(as_uuid=True), nullable=False
    )
    segment_id: Mapped[int] = mapped_column(primary_key=True)
    duration: Mapped[timedelta] = mapped_column(nullable=False)
    pauses: Mapped[Optional[timedelta]] = mapped_column(nullable=True)
    moving: Mapped[Optional[timedelta]] = mapped_column(nullable=True)
    distance: Mapped[Optional[float]] = mapped_column(
        db.Numeric(6, 3), nullable=True
    )  # kilometers
    min_alt: Mapped[Optional[float]] = mapped_column(
        db.Numeric(6, 2), nullable=True
    )  # meters
    max_alt: Mapped[Optional[float]] = mapped_column(
        db.Numeric(6, 2), nullable=True
    )  # meters
    descent: Mapped[Optional[float]] = mapped_column(
        db.Numeric(8, 3), nullable=True
    )  # meters
    ascent: Mapped[Optional[float]] = mapped_column(
        db.Numeric(8, 3), nullable=True
    )  # meters
    max_speed: Mapped[Optional[float]] = mapped_column(
        db.Numeric(6, 2), nullable=True
    )  # km/h
    ave_speed: Mapped[Optional[float]] = mapped_column(
        db.Numeric(6, 2), nullable=True
    )  # km/h

    workout: Mapped["Workout"] = relationship(
        "Workout", lazy="joined", single_parent=True
    )

    def __str__(self) -> str:
        return (
            f"<Segment '{self.segment_id}' "
            f"for workout '{encode_uuid(self.workout_uuid)}'>"
        )

    def __init__(
        self, segment_id: int, workout_id: int, workout_uuid: UUID
    ) -> None:
        self.segment_id = segment_id
        self.workout_id = workout_id
        self.workout_uuid = workout_uuid

    def serialize(self) -> Dict:
        return {
            "workout_id": encode_uuid(self.workout_uuid),
            "segment_id": self.segment_id,
            "duration": str(self.duration) if self.duration else None,
            "pauses": str(self.pauses) if self.pauses else None,
            "moving": str(self.moving) if self.moving else None,
            "distance": float(self.distance) if self.distance else None,
            "min_alt": float(self.min_alt) if self.min_alt else None,
            "max_alt": float(self.max_alt) if self.max_alt else None,
            "descent": float(self.descent) if self.descent else None,
            "ascent": float(self.ascent) if self.ascent else None,
            "max_speed": float(self.max_speed) if self.max_speed else None,
            "ave_speed": float(self.ave_speed) if self.ave_speed else None,
        }


class Record(BaseModel):
    __tablename__ = "records"
    __table_args__ = (
        db.UniqueConstraint(
            "user_id", "sport_id", "record_type", name="user_sports_records"
        ),
        db.Index("workout_records", "workout_id", "record_type"),
    )
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id"), nullable=False
    )
    sport_id: Mapped[int] = mapped_column(
        db.ForeignKey("sports.id"), nullable=False
    )
    workout_id: Mapped[int] = mapped_column(
        db.ForeignKey("workouts.id"), nullable=False
    )
    workout_uuid: Mapped[UUID] = mapped_column(
        postgresql.UUID(as_uuid=True), nullable=False
    )
    record_type: Mapped[str] = mapped_column(
        Enum(*record_types, name="record_types")
    )
    workout_date: Mapped[datetime] = mapped_column(TZDateTime, nullable=False)
    _value: Mapped[Optional[int]] = mapped_column("value", nullable=True)

    user: Mapped["User"] = relationship(
        "User", lazy="joined", single_parent=True
    )
    sport: Mapped["Sport"] = relationship(
        "Sport", lazy="joined", single_parent=True
    )
    workout: Mapped["Workout"] = relationship(
        "Workout", lazy="joined", single_parent=True
    )

    def __str__(self) -> str:
        return (
            f"<Record {self.sport.label} - "
            f"{self.record_type} - "
            f"{self.workout_date.strftime('%Y-%m-%d')}>"
        )

    def __init__(self, workout: Workout, record_type: str) -> None:
        self.user_id = workout.user_id
        self.sport_id = workout.sport_id
        self.workout_id = workout.id
        self.workout_uuid = workout.uuid
        self.record_type = record_type
        self.workout_date = workout.workout_date

    @hybrid_property
    def value(self) -> Optional[Union[timedelta, float]]:
        if self._value is None:
            return None
        if self.record_type == "LD":
            return timedelta(seconds=self._value)
        elif self.record_type in ["AS", "MS"]:
            return float(self._value / 100)
        else:  # 'FD' or 'HA'
            return float(self._value / 1000)

    @value.setter  # type: ignore
    def value(self, val: Union[datetime, float]) -> None:
        self._value = convert_value_to_integer(self.record_type, val)

    def serialize(self) -> Dict:
        if self.value is None:
            value: Union[float, str, None] = None
        elif self.record_type in ["AS", "FD", "HA", "MS"]:
            value = float(self.value)
        else:  # 'LD'
            value = str(self.value)

        return {
            "id": self.id,
            "user": self.user.username,
            "sport_id": self.sport_id,
            "workout_id": encode_uuid(self.workout_uuid),
            "record_type": self.record_type,
            "workout_date": self.workout_date,
            "value": value,
        }


@listens_for(Record, "after_delete")
def on_record_delete(
    mapper: Mapper, connection: Connection, old_record: Record
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        workout = old_record.workout
        new_records = Workout.get_user_workout_records(
            workout.user_id, workout.sport_id
        )
        for record_type, record_data in new_records.items():
            if (
                record_data["record_value"]
                and record_type == old_record.record_type
            ):
                new_record = Record(
                    workout=record_data["workout"], record_type=record_type
                )
                new_record.value = record_data["record_value"]  # type: ignore
                session.add(new_record)


class WorkoutLike(BaseModel):
    __tablename__ = "workout_likes"
    __table_args__ = (
        db.UniqueConstraint(
            "user_id", "workout_id", name="user_id_workout_id_unique"
        ),
    )
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    created_at: Mapped[datetime] = mapped_column(TZDateTime, nullable=False)
    user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id", ondelete="CASCADE"),
        nullable=False,
    )
    workout_id: Mapped[int] = mapped_column(
        db.ForeignKey("workouts.id", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )

    user: Mapped["User"] = relationship("User", lazy=True)
    workout: Mapped["Workout"] = relationship("Workout", lazy=True)

    def __init__(
        self,
        user_id: int,
        workout_id: int,
        created_at: Optional[datetime] = None,
    ) -> None:
        self.user_id = user_id
        self.workout_id = workout_id
        self.created_at = (
            datetime.now(timezone.utc) if created_at is None else created_at
        )


@listens_for(WorkoutLike, "after_insert")
def on_workout_like_insert(
    mapper: Mapper, connection: Connection, new_workout_like: WorkoutLike
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        from fittrackee.users.models import Notification, User

        workout = Workout.query.filter_by(
            id=new_workout_like.workout_id
        ).first()
        if not workout:
            return
        if new_workout_like.user_id != workout.user_id:
            to_user = User.query.filter_by(id=workout.user_id).first()
            if not to_user or not to_user.is_notification_enabled(
                "workout_like"
            ):
                return

            notification = Notification(
                from_user_id=new_workout_like.user_id,
                to_user_id=workout.user_id,
                created_at=new_workout_like.created_at,
                event_type="workout_like",
                event_object_id=workout.id,
            )
            session.add(notification)


@listens_for(WorkoutLike, "after_delete")
def on_workout_like_delete(
    mapper: Mapper, connection: Connection, old_workout_like: WorkoutLike
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        from fittrackee.users.models import Notification

        workout = Workout.query.filter_by(
            id=old_workout_like.workout_id
        ).first()
        if not workout:
            return
        Notification.query.filter_by(
            from_user_id=old_workout_like.user_id,
            to_user_id=workout.user_id,
            event_type="workout_like",
            event_object_id=workout.id,
        ).delete()



"""add virtual cycling as sport type

Revision ID: 5b936821326d
Revises: 84d840ce853b
Create Date: 2022-10-26 17:43:20.114104

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '5b936821326d'
down_revision = '84d840ce853b'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Cycling (Virtual)', True, 1)
        """
    )


def downgrade():
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Cycling (Virtual)';
        """
    )



"""add analysis visibility level

Revision ID: 8a80bec0a410
Revises: 70f12f8c0218
Create Date: 2024-12-23 10:16:23.026455

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql


# revision identifiers, used by Alembic.
revision = '8a80bec0a410'
down_revision = '70f12f8c0218'
branch_labels = None
depends_on = None

visibility_levels = postgresql.ENUM(
    'PUBLIC',
    'FOLLOWERS_AND_REMOTE',  # for a next version, not used for now
    'FOLLOWERS',
    'PRIVATE',
    name='visibility_levels',
)


def upgrade():
    visibility_levels.create(op.get_bind())
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                'analysis_visibility',
                visibility_levels,
                server_default='PRIVATE',
                nullable=True,
            )
        )
    op.execute("UPDATE users SET analysis_visibility = 'PRIVATE';")
    op.alter_column('users', 'analysis_visibility', nullable=False)

    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                'analysis_visibility',
                visibility_levels,
                server_default='PRIVATE',
                nullable=True,
            )
        )
    op.execute("UPDATE workouts SET analysis_visibility = 'PRIVATE';")
    op.alter_column('workouts', 'analysis_visibility', nullable=False)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.drop_column('analysis_visibility')

    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.drop_column('analysis_visibility')

    # ### end Alembic commands ###



"""add Halfbike

Revision ID: 61ecbc76f410
Revises: 78a90b587a9b
Create Date: 2025-04-15 23:03:32.375827

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '61ecbc76f410'
down_revision = '78a90b587a9b'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Halfbike', True, 1)
        """
    )


def downgrade():
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Halfbike';
        """
    )



import re
from typing import TYPE_CHECKING, Optional, Set, Tuple

from sqlalchemy import func

from fittrackee.comments.exceptions import CommentForbiddenException
from fittrackee.utils import decode_short_id
from fittrackee.visibility_levels import can_view

from .models import Comment

if TYPE_CHECKING:
    from fittrackee.users.models import User

MENTION_REGEX = r"(@(<span\s*.*>)?([\w_\-\.]+))(<\/span>)?"
LINK_TEMPLATE = (
    '<a href="{url}" target="_blank" rel="noopener noreferrer">'
    "@<span>{username}</span></a>"
)


def handle_mentions(text: str) -> Tuple[str, Set["User"]]:
    from fittrackee.users.models import User

    mentioned_users: Set["User"] = set()
    for _, _, username, _ in re.findall(re.compile(MENTION_REGEX), text):
        user = User.query.filter(
            func.lower(User.username) == func.lower(username),
        ).first()
        if user:
            mentioned_users.add(user)
            text = text.replace(
                f"@{username}",
                LINK_TEMPLATE.format(
                    url=user.get_user_url(), username=username
                ),
            )
    return text, mentioned_users


def get_comment(comment_short_id: str, auth_user: Optional["User"]) -> Comment:
    workout_comment_uuid = decode_short_id(comment_short_id)
    filters = [Comment.uuid == workout_comment_uuid]
    if auth_user:
        filters.append(
            Comment.user_id.not_in(
                auth_user.get_blocked_user_ids()
                + auth_user.get_blocked_by_user_ids()
            )
        )
    comment = Comment.query.filter(*filters).first()
    if not comment or not can_view(comment, "text_visibility", auth_user):
        raise CommentForbiddenException()
    return comment



class BlockUserException(Exception):
    pass


class FollowRequestAlreadyProcessedError(Exception):
    pass


class FollowRequestAlreadyRejectedError(Exception):
    pass


class InvalidEmailException(Exception):
    pass


class InvalidNotificationTypeException(Exception):
    pass


class InvalidUserRole(Exception):
    def __init__(self) -> None:
        super().__init__("invalid role")


class MissingAdminIdException(Exception):
    pass


class MissingReportIdException(Exception):
    pass


class NotExistingFollowRequestError(Exception):
    pass


class OwnerException(Exception):
    pass


class UserAlreadyReactivatedException(Exception):
    def __init__(self) -> None:
        super().__init__("user account already reactivated")


class UserAlreadySuspendedException(Exception):
    def __init__(self) -> None:
        super().__init__("user account already suspended")


class UserControlsException(Exception):
    pass


class UserCreationException(Exception):
    pass


class UserNotFoundException(Exception):
    pass



"""create Record table

Revision ID: caf0e0dc621a
Revises: b7cfe0c17708
Create Date: 2018-05-11 22:33:08.267488

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'caf0e0dc621a'
down_revision = 'b7cfe0c17708'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('records',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('sport_id', sa.Integer(), nullable=False),
    sa.Column('activity_id', sa.Integer(), nullable=False),
    sa.Column('record_type', sa.Enum('AS', 'FD', 'LD', 'MS', name='record_types'), nullable=True),
    sa.Column('activity_date', sa.DateTime(), nullable=False),
    sa.Column('value', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['activity_id'], ['activities.id'], ),
    sa.ForeignKeyConstraint(['sport_id'], ['sports.id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('user_id', 'sport_id', 'record_type', name='user_sports_records')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('records')
    # ### end Alembic commands ###



import secrets
from datetime import datetime, timezone
from typing import Optional, Tuple

from sqlalchemy import func

from fittrackee import db
from fittrackee.reports.models import ReportAction
from fittrackee.users.constants import (
    ADMINISTRATOR_NOTIFICATION_TYPES,
    MODERATOR_NOTIFICATION_TYPES,
    USER_DATE_FORMAT,
    USER_TIMEZONE,
)
from fittrackee.users.exceptions import (
    InvalidEmailException,
    InvalidUserRole,
    MissingAdminIdException,
    MissingReportIdException,
    OwnerException,
    UserAlreadyReactivatedException,
    UserAlreadySuspendedException,
    UserControlsException,
    UserCreationException,
    UserNotFoundException,
)
from fittrackee.users.models import (
    Notification,
    User,
)
from fittrackee.users.roles import UserRole
from fittrackee.users.utils.controls import is_valid_email, register_controls


class UserManagerService:
    def __init__(self, username: str, moderator_id: Optional[int] = None):
        self.username = username
        self.moderator_id = moderator_id

    def _get_user(self) -> User:
        user = User.query.filter_by(username=self.username).first()
        if not user:
            raise UserNotFoundException()
        return user

    @staticmethod
    def _update_active_status(user: User, active_status: bool) -> None:
        user.is_active = active_status
        if active_status:
            user.confirmation_token = None

    @staticmethod
    def _reset_user_password(user: User) -> str:
        new_password = secrets.token_urlsafe(30)
        user.password = user.generate_password_hash(new_password)
        return new_password

    @staticmethod
    def _update_user_email(
        user: User, new_email: str, with_confirmation: bool
    ) -> None:
        if not is_valid_email(new_email):
            raise InvalidEmailException("valid email must be provided")
        if user.email == new_email:
            raise InvalidEmailException(
                "new email must be different than current email"
            )
        if with_confirmation:
            user.email_to_confirm = new_email
            user.confirmation_token = secrets.token_urlsafe(30)
        else:
            user.email = new_email

    def update(
        self,
        role: Optional[str] = None,
        activate: Optional[bool] = None,
        reset_password: bool = False,
        new_email: Optional[str] = None,
        with_confirmation: bool = True,
        suspended: Optional[bool] = None,
        report_id: Optional[int] = None,
        reason: Optional[str] = None,
        raise_error_on_owner: bool = False,
    ) -> Tuple[User, bool, Optional[str], Optional[ReportAction]]:
        user_updated = False
        new_password = None
        report_action = None
        user = self._get_user()

        if user.role == UserRole.OWNER.value and raise_error_on_owner:
            raise OwnerException("user with owner rights can not be modified")

        if suspended is not None:
            if self.moderator_id is None:
                raise MissingAdminIdException()
            if report_id is None:
                raise MissingReportIdException()

        if role is not None:
            if role not in UserRole.db_choices():
                raise InvalidUserRole()
            user.role = UserRole[role.upper()].value
            if user.role >= UserRole.MODERATOR.value:
                activate = True
            if user.role < UserRole.MODERATOR.value:
                db.session.query(Notification).filter(
                    Notification.to_user_id == user.id,
                    Notification.event_type.in_(MODERATOR_NOTIFICATION_TYPES),
                ).delete()
            if user.role < UserRole.ADMIN.value:
                db.session.query(Notification).filter(
                    Notification.to_user_id == user.id,
                    Notification.event_type.in_(
                        ADMINISTRATOR_NOTIFICATION_TYPES
                    ),
                ).delete()
            user_updated = True

        if activate is not None:
            self._update_active_status(user, activate)
            user_updated = True

        if reset_password:
            new_password = self._reset_user_password(user)
            user_updated = True

        if new_email is not None:
            self._update_user_email(user, new_email, with_confirmation)
            user_updated = True

        now = datetime.now(timezone.utc)
        if suspended is True:
            if user.suspended_at:
                raise UserAlreadySuspendedException()
            user.suspended_at = now
            user.role = UserRole.USER.value
            user_updated = True
        if suspended is False:
            if user.suspended_at is None:
                raise UserAlreadyReactivatedException()
            user.suspended_at = None
            user_updated = True
        if self.moderator_id and report_id and suspended is not None:
            report_action = ReportAction(
                moderator_id=self.moderator_id,
                action_type=(
                    "user_suspension" if suspended else "user_unsuspension"
                ),
                created_at=now,
                report_id=report_id,
                reason=reason,
                user_id=user.id,
            )
            db.session.add(report_action)

        db.session.commit()
        return user, user_updated, new_password, report_action

    def create_user(
        self,
        email: str,
        password: Optional[str] = None,
        check_email: bool = False,
        role: Optional[str] = None,
    ) -> Tuple[Optional[User], Optional[str]]:
        if not password:
            password = secrets.token_urlsafe(30)

        ret = register_controls(self.username, email, password)

        if ret != "":
            raise UserControlsException(ret)

        user = User.query.filter(
            func.lower(User.username) == func.lower(self.username)
        ).first()
        if user:
            raise UserCreationException(
                "sorry, that username is already taken"
            )

        # if a user exists with same email address, no error is returned
        # since a user has to confirm his email to activate his account
        user = User.query.filter(
            func.lower(User.email) == func.lower(email)
        ).first()
        if user:
            if check_email:
                raise UserCreationException(
                    "This user already exists. No action done."
                )
            return None, None

        new_user = User(username=self.username, email=email, password=password)
        new_user.timezone = USER_TIMEZONE
        new_user.date_format = USER_DATE_FORMAT
        new_user.confirmation_token = secrets.token_urlsafe(30)

        if role is not None:
            if role not in UserRole.db_choices():
                raise InvalidUserRole()
            new_user.role = UserRole[role.upper()].value

        db.session.add(new_user)
        db.session.flush()

        return new_user, password

    def create(
        self,
        email: str,
        password: Optional[str] = None,
        role: Optional[str] = None,
    ) -> Tuple[Optional[User], Optional[str]]:
        try:
            new_user, password = self.create_user(
                email, password, check_email=True, role=role
            )
            if new_user:
                new_user.language = "en"
                db.session.commit()
        except UserControlsException as e:
            raise UserCreationException(str(e)) from e
        return new_user, password



"""add Kayaking and Canoeing

Revision ID: 78a90b587a9b
Revises: ce68b3914ff7
Create Date: 2025-03-02 16:08:56.898458

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '78a90b587a9b'
down_revision = 'ce68b3914ff7'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Kayaking', True, 1),
               ('Canoeing', True, 1)
        """
    )


def downgrade():
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Kayaking' OR label = 'Canoeing';
        """
    )



from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Optional, Tuple, Union

import gpxpy.gpx

from ..exceptions import InvalidGPXException, WorkoutGPXException
from .weather import WeatherService

weather_service = WeatherService()


def open_gpx_file(gpx_file: str) -> Optional[gpxpy.gpx.GPX]:
    gpx_file = open(gpx_file, "r")  # type: ignore
    gpx = gpxpy.parse(gpx_file)
    if len(gpx.tracks) == 0:
        return None
    return gpx


def get_gpx_data(
    parsed_gpx: Union[gpxpy.gpx.GPX, gpxpy.gpx.GPXTrackSegment],
    max_speed: float,
    start: Union[datetime, None],
    stopped_time_between_seg: timedelta,
    stopped_speed_threshold: float,
) -> Dict:
    """
    Returns data from parsed gpx file
    """
    gpx_data: Dict[str, Any] = {
        "max_speed": (max_speed / 1000) * 3600,
        "start": start,
    }

    duration = parsed_gpx.get_duration()
    gpx_data["duration"] = (
        timedelta(seconds=duration if duration else 0)
        + stopped_time_between_seg
    )

    ele = parsed_gpx.get_elevation_extremes()
    gpx_data["elevation_max"] = ele.maximum
    gpx_data["elevation_min"] = ele.minimum

    # gpx file contains elevation data (<ele> element)
    if ele.maximum is not None:
        hill = parsed_gpx.get_uphill_downhill()
        gpx_data["uphill"] = hill.uphill
        gpx_data["downhill"] = hill.downhill
    else:
        gpx_data["uphill"] = None
        gpx_data["downhill"] = None

    moving_data = parsed_gpx.get_moving_data(
        stopped_speed_threshold=stopped_speed_threshold
    )
    if moving_data:
        gpx_data["moving_time"] = timedelta(seconds=moving_data.moving_time)
        gpx_data["stop_time"] = (
            timedelta(seconds=moving_data.stopped_time)
            + stopped_time_between_seg
        )
        distance = moving_data.moving_distance + moving_data.stopped_distance
        gpx_data["distance"] = distance / 1000

        average_speed = (
            distance / moving_data.moving_time
            if moving_data.moving_time > 0
            else 0
        )
        gpx_data["average_speed"] = (average_speed / 1000) * 3600

    return gpx_data


def get_gpx_info(
    gpx_file: str,
    stopped_speed_threshold: float,
    update_map_data: Optional[bool] = True,
    update_weather_data: Optional[bool] = True,
    use_raw_gpx_speed: bool = False,
) -> Tuple:
    """
    Parse and return gpx, map and weather data from gpx file
    """
    try:
        gpx = open_gpx_file(gpx_file)
    except Exception as e:
        raise InvalidGPXException("error", "gpx file is invalid") from e
    if gpx is None:
        raise InvalidGPXException("error", "no tracks in gpx file")

    gpx_data: Dict = {
        "name": gpx.tracks[0].name,
        "description": gpx.tracks[0].description,
        "segments": [],
    }
    max_speed = 0.0
    start: Optional[datetime] = None
    map_data = []
    weather_data = []
    segments_nb = len(gpx.tracks[0].segments)
    prev_seg_last_point = None
    no_stopped_time = timedelta(seconds=0)
    stopped_time_between_seg = no_stopped_time

    for segment_idx, segment in enumerate(gpx.tracks[0].segments):
        segment_start: Optional[datetime] = None
        segment_points_nb = len(segment.points)
        for point_idx, point in enumerate(segment.points):
            if point.time is None:
                raise InvalidGPXException(
                    "error", "<time> is missing in gpx file"
                )
            if point_idx == 0:
                segment_start = point.time
                # first gpx point => get weather
                if start is None:
                    start = point.time
                    if point.time and update_weather_data:
                        weather_data.append(weather_service.get_weather(point))

                # if a previous segment exists, calculate stopped time between
                # the two segments
                if prev_seg_last_point:
                    stopped_time_between_seg += (
                        point.time - prev_seg_last_point
                    )

            # last segment point
            if point_idx == (segment_points_nb - 1):
                prev_seg_last_point = point.time

                # last gpx point => get weather
                if segment_idx == (segments_nb - 1) and update_weather_data:
                    weather_data.append(weather_service.get_weather(point))

            if update_map_data:
                map_data.append([point.longitude, point.latitude])
        moving_data = segment.get_moving_data(
            stopped_speed_threshold=stopped_speed_threshold,
            raw=use_raw_gpx_speed,
        )
        if moving_data:
            calculated_max_speed = moving_data.max_speed
            segment_max_speed = (
                calculated_max_speed if calculated_max_speed else 0
            )

            if segment_max_speed > max_speed:
                max_speed = segment_max_speed
        else:
            segment_max_speed = 0.0

        segment_data = get_gpx_data(
            segment,
            segment_max_speed,
            segment_start,
            no_stopped_time,
            stopped_speed_threshold,
        )
        segment_data["idx"] = segment_idx
        gpx_data["segments"].append(segment_data)

    full_gpx_data = get_gpx_data(
        gpx,
        max_speed,
        start,
        stopped_time_between_seg,
        stopped_speed_threshold,
    )
    gpx_data = {**gpx_data, **full_gpx_data}

    if update_map_data:
        bounds = gpx.get_bounds()
        gpx_data["bounds"] = (
            [
                bounds.min_latitude,
                bounds.min_longitude,
                bounds.max_latitude,
                bounds.max_longitude,
            ]
            if bounds
            else []
        )

    return gpx_data, map_data, weather_data


def get_gpx_segments(
    track_segments: List, segment_id: Optional[int] = None
) -> List:
    """
    Return list of segments, filtered on segment id if provided
    """
    if segment_id is not None:
        segment_index = segment_id - 1
        if segment_index > (len(track_segments) - 1):
            raise WorkoutGPXException(
                "not found", f"No segment with id '{segment_id}'", None
            )
        if segment_index < 0:
            raise WorkoutGPXException("error", "Incorrect segment id", None)
        segments = [track_segments[segment_index]]
    else:
        segments = track_segments

    return segments


def get_chart_data(
    gpx_file: str, segment_id: Optional[int] = None
) -> Optional[List]:
    """
    Return data needed to generate chart with speed and elevation
    """
    gpx = open_gpx_file(gpx_file)
    if gpx is None:
        return None

    chart_data = []
    first_point = None
    previous_point = None
    previous_distance = 0

    track_segments = gpx.tracks[0].segments
    segments = get_gpx_segments(track_segments, segment_id)

    for segment_idx, segment in enumerate(segments):
        for point_idx, point in enumerate(segment.points):
            if segment_idx == 0 and point_idx == 0:
                first_point = point
            distance = (
                point.distance_3d(previous_point)
                if (
                    point.elevation
                    and previous_point
                    and previous_point.elevation
                )
                else point.distance_2d(previous_point)
            )
            distance = 0 if distance is None else distance
            distance += previous_distance
            speed = (
                round((segment.get_speed(point_idx) / 1000) * 3600, 2)
                if segment.get_speed(point_idx) is not None
                else 0
            )
            data = {
                "distance": (
                    round(distance / 1000, 2) if distance is not None else 0
                ),
                "duration": int(point.time_difference(first_point)),
                "latitude": point.latitude,
                "longitude": point.longitude,
                "speed": speed,
                # workaround
                # https://github.com/tkrajina/gpxpy/issues/209
                "time": point.time.replace(
                    tzinfo=timezone(point.time.utcoffset())
                ),
            }
            if point.elevation:
                data["elevation"] = round(point.elevation, 1)
            chart_data.append(data)
            previous_point = point
            previous_distance = distance

    return chart_data


def extract_segment_from_gpx_file(
    content: str, segment_id: int
) -> Optional[str]:
    """
    Returns segments in xml format from a gpx file content
    """
    gpx_content = gpxpy.parse(content)
    if len(gpx_content.tracks) == 0:
        return None

    track_segment = get_gpx_segments(
        gpx_content.tracks[0].segments, segment_id
    )

    gpx = gpxpy.gpx.GPX()
    gpx_track = gpxpy.gpx.GPXTrack()
    gpx.tracks.append(gpx_track)
    gpx_segment = gpxpy.gpx.GPXTrackSegment()
    gpx_track.segments.append(gpx_segment)

    for point in track_segment[0].points:
        gpx_segment.points.append(
            gpxpy.gpx.GPXTrackPoint(
                point.latitude, point.longitude, elevation=point.elevation
            )
        )

    return gpx.to_xml()



"""add paragliding sport

Revision ID: d22670a89a54
Revises: eff1c16c43eb
Create Date: 2023-07-12 21:00:11.291517

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd22670a89a54'
down_revision = 'eff1c16c43eb'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Paragliding', True, 0.1)
        """
    )

def downgrade():
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Paragliding';
        """
    )



"""add 'timezone' to 'User' table

Revision ID: e82e5e9447de
Revises: 9f8c9c37da44
Create Date: 2018-06-08 16:01:52.684935

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e82e5e9447de'
down_revision = '9f8c9c37da44'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('users', sa.Column('timezone', sa.String(length=50), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('users', 'timezone')
    # ### end Alembic commands ###



from datetime import timedelta
from typing import Dict, List, Union

from flask import Blueprint, current_app, request
from sqlalchemy import func

from fittrackee import db
from fittrackee.dates import get_datetime_in_utc
from fittrackee.oauth2.server import require_auth
from fittrackee.responses import (
    ForbiddenErrorResponse,
    HttpResponse,
    InvalidPayloadErrorResponse,
    NotFoundErrorResponse,
    UserNotFoundErrorResponse,
    handle_error_and_return_response,
)
from fittrackee.users.models import User
from fittrackee.users.roles import UserRole

from .models import Sport, Workout
from .utils.uploads import get_upload_dir_size
from .utils.workouts import get_average_speed, get_datetime_from_request_args

stats_blueprint = Blueprint("stats", __name__)


def get_stats_from_row(row: List, stats_type: str) -> Dict:
    row_stats = {
        "total_workouts": row[2],
        f"{stats_type}_distance": round(float(row[3]), 2),
        f"{stats_type}_duration": int(row[4].total_seconds()),
        f"{stats_type}_ascent": (
            None if row[5] is None else round(float(row[5]), 2)
        ),
        f"{stats_type}_descent": (
            None if row[6] is None else round(float(row[6]), 2)
        ),
    }
    if stats_type == "average":
        row_stats["average_speed"] = round(float(row[1]), 2)
    return row_stats


@stats_blueprint.route("/stats/<user_name>/by_time", methods=["GET"])
@require_auth(scopes=["workouts:read"])
def get_workouts_by_time(
    auth_user: User, user_name: str
) -> Union[Dict, HttpResponse]:
    """
    Get workouts statistics for a user by time.
    For now only authenticated users can access their statistics.

    **Scope**: ``workouts:read``

    **Example requests**:

    - without parameters:

    .. sourcecode:: http

      GET /api/stats/admin/by_time HTTP/1.1

    - with parameters:

    .. sourcecode:: http

      GET /api/stats/admin/by_time?from=2018-01-01&to=2018-06-30&time=week
        HTTP/1.1

    **Example responses**:

    - success for total:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "statistics": {
            "2017": {
              "3": {
                "total_workouts": 2,
                "total_ascent": 203.0,
                "total_ascent": 156.0,
                "total_distance": 15.282,
                "total_duration": 12341
              }
            },
            "2019": {
              "1": {
                "total_workouts": 3,
                "total_ascent": 150.0,
                "total_ascent": 178.0,
                "total_distance": 47,
                "total_duration": 9960
              },
              "2": {
                "total_workouts": 1,
                "total_ascent": 46.0,
                "total_ascent": 78.0,
                "total_distance": 5.613,
                "total_duration": 1267
              }
            }
          }
        },
        "status": "success"
      }

    - success for average:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "statistics": {
            "2017": {
              "3": {
                "average_ascent": 101.5,
                "average_ascent": 78.0,
                "average_distance": 15.282,
                "average_duration": 7641,
                "average_speed": 4.48,
                "total_workouts": 2
              }
            },
            "2019": {
              "1": {
                "average_ascent": 50.0,
                "average_descent": 59.33,
                "average_distance": 15.67,
                "average_duration": 3320,
                "average_speed": 16.99,
                "total_workouts": 3
              },
              "2": {
                "average_ascent": 46.0,
                "average_descent": 78.0,
                "average_distance": 5.613,
                "average_duration": 1267,
                "average_speed": 15.95,
                "total_workouts": 1
              }
            }
          }
        },
        "status": "success"
      }

    - no workouts:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
            "statistics": {}
        },
        "status": "success"
      }

    :param integer user_name: username

    :query string from: start date (format: ``%Y-%m-%d``)
    :query string to: end date (format: ``%Y-%m-%d``)
    :query string time: time frame:

      - ``day``: day
      - ``week``: week starting Sunday
      - ``weekm``: week starting Monday
      - ``month``: month
      - ``year``: year (default)
    :query string type: stats type:

      - ``total``: calculating totals
      - ``average``: calculating averages

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid stats type``
        - ``invalid time period``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user does not exist``

    """
    try:
        user = User.query.filter_by(username=user_name).first()
        if not user:
            return UserNotFoundErrorResponse()
        if user.id != auth_user.id:
            return ForbiddenErrorResponse()

        params = request.args.copy()
        date_from, date_to = get_datetime_from_request_args(params, user)
        time = params.get("time")
        stats_type = params.get("type", "total")
        if stats_type not in ["total", "average"]:
            return InvalidPayloadErrorResponse("invalid stats type", "fail")

        if not time or time == "year":
            time_format = "yyyy"
        elif time == "day":
            time_format = "yyyy-mm-dd"
        elif time == "month":
            time_format = "yyyy-mm"
        elif time.startswith("week"):
            # 'week' => week starts on Sunday
            # 'weekm' => week starts on Monday
            #
            # Note: on PostgreSQL, week starts on Monday
            time_format = "IYYY-IW"
        else:
            return InvalidPayloadErrorResponse("invalid time period", "fail")

        # On PostgreSQL, week starts on Monday
        # For 'week' timeframe, the workaround is to add 1 day
        delta = timedelta(days=1 if time and time == "week" else 0)

        calculation_method = func.avg if stats_type == "average" else func.sum
        stats_key = func.to_char(
            func.timezone(
                # user has always timezone set
                auth_user.timezone if auth_user.timezone else "UTC",
                # workout date is stored without timezone in database
                func.timezone("Z", Workout.workout_date),
            )
            + delta,
            time_format,
        )
        filters = [Workout.user_id == user.id]
        if date_from:
            filters.append(Workout.workout_date >= date_from)
        if date_to:
            filters.append(
                Workout.workout_date < date_to + timedelta(seconds=1)
            )

        workouts_query = (
            db.session.query(
                Workout.sport_id,
                (
                    func.avg(Workout.ave_speed)  # type: ignore
                    if stats_type == "average"
                    else True
                ),
                func.count(Workout.id),
                calculation_method(Workout.distance),
                calculation_method(Workout.moving),
                calculation_method(Workout.ascent),
                calculation_method(Workout.descent),
                stats_key,
            )
            .filter(*filters)
            .group_by(stats_key, Workout.sport_id)
        )

        if current_app.config["stats_workouts_limit"]:
            last_workout_ids = (
                db.session.query(
                    Workout.id,
                )
                .order_by(Workout.workout_date.desc())
                .limit(current_app.config["stats_workouts_limit"])
                .subquery()
            )
            workouts_query = workouts_query.join(
                last_workout_ids, Workout.id == last_workout_ids.c.id
            )
        results = workouts_query.all()

        statistics = {}
        for row in results:
            date_key = row[7]
            if time and time.startswith("week"):
                date_key = (
                    get_datetime_in_utc(date_key + "-1", "%G-%V-%u") - delta
                ).strftime("%Y-%m-%d")
            sport_key = row[0]
            if date_key not in statistics:
                statistics[date_key] = {
                    sport_key: get_stats_from_row(list(row), stats_type)
                }
            elif sport_key not in statistics[date_key]:
                statistics[date_key][sport_key] = get_stats_from_row(
                    list(row), stats_type
                )
            else:
                statistics[date_key][sport_key]["total_workouts"] += row[2]
                if stats_type == "average":
                    statistics[date_key][sport_key]["average_speed"] = (
                        get_average_speed(
                            statistics[date_key][sport_key]["total_workouts"],
                            statistics[date_key][sport_key]["average_speed"],
                            row[1],
                        )
                    )
                statistics[date_key][sport_key][f"{stats_type}_distance"] += (
                    round(float(row[3]), 2)
                )
                statistics[date_key][sport_key][f"{stats_type}_duration"] += (
                    int(row[4].total_seconds())
                )
                if row[5]:
                    statistics[date_key][sport_key][
                        f"{stats_type}_ascent"
                    ] += round(float(row[5]), 2)
                if row[6]:
                    statistics[date_key][sport_key][
                        f"{stats_type}_ascent"
                    ] += round(float(row[6]), 2)

        return {
            "status": "success",
            "data": {"statistics": statistics},
        }
    except Exception as e:
        return handle_error_and_return_response(e)


@stats_blueprint.route("/stats/<user_name>/by_sport", methods=["GET"])
@require_auth(scopes=["workouts:read"])
def get_workouts_by_sport(
    auth_user: User, user_name: str
) -> Union[Dict, HttpResponse]:
    """
    Get workouts statistics for a user by sport.
    For now only authenticated users can access their statistics.

    **Scope**: ``workouts:read``

    **Example requests**:

    - without parameters (get stats for all sports with workouts):

    .. sourcecode:: http

      GET /api/stats/admin/by_sport HTTP/1.1

    - with sport id:

    .. sourcecode:: http

      GET /api/stats/admin/by_sport?sport_id=1 HTTP/1.1

    **Example responses**:

    - success for all sports:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "statistics": {
            "1": {
              "average_ascent": 50.0,
              "average_descent": 59.33,
              "average_distance": 15.67,
              "average_duration": 3320,
              "average_speed": 16.99,
              "total_workouts": 3,
              "total_ascent": 150.0,
              "total_ascent": 178.0,
              "total_distance": 47,
              "total_duration": 9960
            },
            "2": {
              "average_ascent": 46.0,
              "average_descent": 78.0,
              "average_distance": 5.613,
              "average_duration": 1267,
              "average_speed": 15.95,
              "total_workouts": 1,
              "total_ascent": 46.0,
              "total_ascent": 78.0,
              "total_distance": 5.613,
              "total_duration": 1267
            },
            "3": {
              "average_ascent": 101.5,
              "average_ascent": 78.0,
              "average_distance": 15.282,
              "average_duration": 7641,
              "average_speed": 4.48,
              "total_workouts": 2,
              "total_ascent": 203.0,
              "total_ascent": 156.0,
              "total_distance": 15.282,
              "total_duration": 12341
            }
          }
        },
        "status": "success"
      }

    - success for a given sport:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "statistics": {
            "2": {
              "average_ascent": 46.0,
              "average_descent": 78.0,
              "average_distance": 5.613,
              "average_duration": 1267,
              "average_speed": 15.95,
              "total_workouts": 1,
              "total_ascent": 46.0,
              "total_ascent": 78.0,
              "total_distance": 5.613,
              "total_duration": 1267
            }
          }
        },
        "status": "success"
      }

    - no workouts:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
            "statistics": {}
        },
        "status": "success"
      }

    :param integer user_name: username

    :query integer sport_id: sport id (not mandatory).
           If not provided, statistics for all sports are returned.

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid stats type``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user does not exist``
        - ``sport does not exist``

    """
    try:
        user = User.query.filter_by(username=user_name).first()
        if not user:
            return UserNotFoundErrorResponse()
        if user.id != auth_user.id:
            return ForbiddenErrorResponse()

        params = request.args.copy()
        sport_id = params.get("sport_id")
        if sport_id:
            sport = Sport.query.filter_by(id=sport_id).first()
            if not sport:
                return NotFoundErrorResponse("sport does not exist")

        filters = [Workout.user_id == user.id]
        if sport_id:
            filters.append(Workout.sport_id == sport_id)
        workouts_query = Workout.query.filter(*filters)
        total_workouts = workouts_query.count()

        workouts_query = workouts_query.order_by(Workout.workout_date.desc())
        if current_app.config["stats_workouts_limit"]:
            workouts_query = workouts_query.limit(
                current_app.config["stats_workouts_limit"]
            )
        workouts_subquery = workouts_query.subquery()
        results = (
            db.session.query(
                workouts_subquery.c.sport_id,
                func.avg(workouts_subquery.c.ave_speed),
                func.avg(workouts_subquery.c.ascent),
                func.avg(workouts_subquery.c.descent),
                func.avg(workouts_subquery.c.distance),
                func.avg(workouts_subquery.c.moving),
                func.sum(workouts_subquery.c.ascent),
                func.sum(workouts_subquery.c.descent),
                func.sum(workouts_subquery.c.distance),
                func.sum(workouts_subquery.c.moving),
                func.count(workouts_subquery.c.id),
            )
            .group_by(workouts_subquery.c.sport_id)
            .all()
        )

        statistics = {}
        for row in results:
            statistics[row[0]] = {
                "average_speed": round(float(row[1]), 2),
                "average_ascent": (
                    None if row[2] is None else round(float(row[2]), 2)
                ),
                "average_descent": (
                    None if row[3] is None else round(float(row[3]), 2)
                ),
                "average_distance": round(float(row[4]), 2),
                "average_duration": str(row[5]).split(".")[0],
                "total_ascent": (
                    None if row[6] is None else round(float(row[6]), 2)
                ),
                "total_descent": (
                    None if row[7] is None else round(float(row[7]), 2)
                ),
                "total_distance": round(float(row[8]), 2),
                "total_duration": str(row[9]).split(".")[0],
                "total_workouts": row[10],
            }

        return {
            "status": "success",
            "data": {
                "statistics": statistics,
                "total_workouts": total_workouts,
            },
        }
    except Exception as e:
        return handle_error_and_return_response(e)


@stats_blueprint.route("/stats/all", methods=["GET"])
@require_auth(scopes=["workouts:read"], role=UserRole.MODERATOR)
def get_application_stats(auth_user: User) -> Dict:
    """
    Get all application statistics.

    **Scope**: ``workouts:read``

    **Minimum role**: Moderator

    **Example requests**:

    .. sourcecode:: http

      GET /api/stats/all HTTP/1.1


    **Example responses**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "sports": 3,
          "uploads_dir_size": 1000,
          "users": 2,
          "workouts": 3,
        },
        "status": "success"
      }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    """

    total_workouts = Workout.query.filter().count()
    nb_users = User.query.filter(User.is_active == True).count()  # noqa
    nb_sports = (
        db.session.query(func.count(Workout.sport_id))
        .group_by(Workout.sport_id)
        .count()
    )
    return {
        "status": "success",
        "data": {
            "workouts": total_workouts,
            "sports": nb_sports,
            "users": nb_users,
            "uploads_dir_size": get_upload_dir_size(),
        },
    }



import json
import os
import secrets
from datetime import datetime, timedelta, timezone
from typing import Dict, List, Optional, Tuple, Union
from zipfile import ZipFile

from flask import current_app

from fittrackee import appLog, db
from fittrackee.emails.tasks import send_email
from fittrackee.files import get_absolute_file_path

from .models import User, UserDataExport
from .utils.language import get_language


class UserDataExporter:
    """
    generates a zip archive with:
    - user info from database (json file)
    - data from database for all workouts if exist (json file)
    - profile picture file if exists
    - gpx files if exist
    """

    def __init__(self, user: User) -> None:
        self.user = user
        self.export_directory = get_absolute_file_path(
            os.path.join("exports", str(self.user.id))
        )
        os.makedirs(self.export_directory, exist_ok=True)
        self.workouts_directory = get_absolute_file_path(
            os.path.join("workouts", str(self.user.id))
        )

    def get_user_info(self) -> Dict:
        return self.user.serialize(current_user=self.user)

    def get_user_workouts_data(self) -> List[Dict]:
        workouts_data = []
        for workout in self.user.workouts:
            workout_data = workout.get_workout_data(
                self.user, additional_data=True, light=False
            )
            workout_data["sport_label"] = workout.sport.label
            workout_data["gpx"] = (
                workout.gpx.split("/")[-1] if workout.gpx else None
            )
            workouts_data.append(workout_data)
        return workouts_data

    def get_user_comments_data(self) -> List[Dict]:
        comments_data = []
        for comment in self.user.comments:
            comments_data.append(
                {
                    "created_at": comment.created_at,
                    "id": comment.short_id,
                    "modification_date": comment.modification_date,
                    "text": comment.text,
                    "text_visibility": comment.text_visibility.value,
                    "workout_id": (
                        comment.workout.short_id
                        if comment.workout_id
                        else None
                    ),
                }
            )
        return comments_data

    def get_user_equipments_data(self) -> List[Dict]:
        return [
            equipment.serialize(current_user=self.user)
            for equipment in self.user.equipments
        ]

    def export_data(self, data: Union[Dict, List], name: str) -> str:
        """export data in existing user upload directory"""
        json_object = json.dumps(data, indent=4, default=str)
        file_path = os.path.join(self.export_directory, f"{name}.json")
        with open(file_path, "w") as export_file:
            export_file.write(json_object)
        return file_path

    def generate_archive(self) -> Tuple[Optional[str], Optional[str]]:
        try:
            user_data_file_name = self.export_data(
                self.get_user_info(), "user_data"
            )
            workout_data_file_name = self.export_data(
                self.get_user_workouts_data(), "workouts_data"
            )
            equipments_data_file_name = self.export_data(
                self.get_user_equipments_data(), "equipments_data"
            )
            comments_data_file_name = self.export_data(
                self.get_user_comments_data(), "comments_data"
            )
            zip_file = f"archive_{secrets.token_urlsafe(15)}.zip"
            zip_path = os.path.join(self.export_directory, zip_file)
            with ZipFile(zip_path, "w") as zip_object:
                zip_object.write(user_data_file_name, "user_data.json")
                zip_object.write(
                    workout_data_file_name, "user_workouts_data.json"
                )
                zip_object.write(
                    equipments_data_file_name, "user_equipments_data.json"
                )
                zip_object.write(
                    comments_data_file_name, "user_comments_data.json"
                )
                if self.user.picture:
                    picture_path = get_absolute_file_path(self.user.picture)
                    if os.path.isfile(picture_path):
                        zip_object.write(
                            picture_path, self.user.picture.split("/")[-1]
                        )
                if os.path.exists(self.workouts_directory):
                    for file in os.listdir(self.workouts_directory):
                        if os.path.isfile(
                            os.path.join(self.workouts_directory, file)
                        ) and file.endswith(".gpx"):
                            zip_object.write(
                                os.path.join(self.workouts_directory, file),
                                f"gpx/{file}",
                            )

            file_exists = os.path.exists(zip_path)
            os.remove(user_data_file_name)
            os.remove(workout_data_file_name)
            os.remove(equipments_data_file_name)
            os.remove(comments_data_file_name)
            return (zip_path, zip_file) if file_exists else (None, None)
        except Exception as e:
            appLog.error(f"Error when generating user data archive: {e!s}")
            return None, None


def export_user_data(export_request_id: int) -> None:
    export_request = UserDataExport.query.filter_by(
        id=export_request_id
    ).first()

    if not export_request:
        appLog.error(f"No export to process for id '{export_request_id}'")
        return

    if export_request.completed:
        appLog.info(f"Export id '{export_request_id}' already processed")
        return

    user = User.query.filter_by(id=export_request.user_id).one()
    exporter = UserDataExporter(user)
    archive_file_path, archive_file_name = exporter.generate_archive()

    try:
        export_request.completed = True
        if archive_file_name and archive_file_path:
            export_request.file_name = archive_file_name
            export_request.file_size = os.path.getsize(archive_file_path)
            db.session.commit()

            if current_app.config["CAN_SEND_EMAILS"]:
                fittrackee_url = current_app.config["UI_URL"]
                email_data = {
                    "username": user.username,
                    "fittrackee_url": fittrackee_url,
                    "account_url": f"{fittrackee_url}/profile/edit/account",
                }
                user_data = {
                    "language": get_language(user.language),
                    "email": user.email,
                }
                send_email.send(
                    user_data, email_data, template="data_export_ready"
                )
        else:
            db.session.commit()
    except Exception as e:
        appLog.error(f"Error when exporting user data: {e!s}")


def clean_user_data_export(days: int) -> Dict:
    counts = {"deleted_requests": 0, "deleted_archives": 0, "freed_space": 0}
    limit = datetime.now(timezone.utc) - timedelta(days=days)
    export_requests = UserDataExport.query.filter(
        UserDataExport.created_at < limit,
        UserDataExport.completed == True,  # noqa
    ).all()

    if not export_requests:
        return counts

    archive_directory = get_absolute_file_path("exports")
    for request in export_requests:
        if request.file_name:
            archive_path = os.path.join(
                archive_directory, f"{request.user_id}", request.file_name
            )
            if os.path.exists(archive_path):
                counts["deleted_archives"] += 1
                counts["freed_space"] += request.file_size
        # Archive is deleted when row is deleted
        db.session.delete(request)
        counts["deleted_requests"] += 1

    db.session.commit()
    return counts


def generate_user_data_archives(max_count: int) -> int:
    count = 0
    export_requests = (
        db.session.query(UserDataExport)
        .filter(UserDataExport.completed == False)  # noqa
        .order_by(UserDataExport.created_at)
        .limit(max_count)
        .all()
    )

    for export_request in export_requests:
        export_user_data(export_request.id)
        count += 1

    return count



from datetime import datetime, timezone
from typing import Dict, Optional, Union

from sqlalchemy import func
from sqlalchemy.orm.exc import NoResultFound

from fittrackee import db
from fittrackee.comments.models import Comment
from fittrackee.comments.utils import get_comment
from fittrackee.users.exceptions import UserNotFoundException
from fittrackee.users.models import User
from fittrackee.users.users_service import UserManagerService
from fittrackee.workouts.models import Workout
from fittrackee.workouts.utils.workouts import get_workout

from .exceptions import (
    InvalidReportActionException,
    InvalidReportException,
    ReportNotFoundException,
    SuspendedObjectException,
    UserWarningExistsException,
)
from .models import (
    ALL_USER_ACTION_TYPES,
    COMMENT_ACTION_TYPES,
    WORKOUT_ACTION_TYPES,
    Report,
    ReportAction,
    ReportActionAppeal,
    ReportComment,
)


class ReportService:
    @staticmethod
    def create_report(
        *,
        reporter: User,
        note: str,
        object_id: str,
        object_type: str,
    ) -> Report:
        target_object: Union[Comment, User, Workout]
        if object_type == "comment":
            target_object = get_comment(object_id, reporter)
        elif object_type == "workout":
            target_object = get_workout(object_id, reporter)
        else:  # object_type == "user"
            try:
                user = User.query.filter(
                    func.lower(User.username) == func.lower(object_id),
                ).one()
            except NoResultFound as e:
                raise UserNotFoundException() from e
            if not user.is_active:
                raise UserNotFoundException()
            target_object = user

        if target_object and target_object.suspended_at:
            raise SuspendedObjectException(object_type)

        existing_unresolved_report = Report.query.filter_by(
            reported_by=reporter.id,
            resolved=False,
            object_type=object_type,
            **{f"reported_{object_type}_id": target_object.id},
        ).first()
        if existing_unresolved_report:
            raise InvalidReportException("a report already exists")

        new_report = Report(
            note=note,
            reported_by=reporter.id,
            reported_object=target_object,
        )
        db.session.add(new_report)
        db.session.commit()

        return new_report

    @staticmethod
    def update_report(
        *,
        report_id: int,
        moderator: User,
        report_comment: str,
        resolved: Optional[bool] = None,
    ) -> Report:
        report = Report.query.filter_by(id=report_id).first()
        if not report:
            raise ReportNotFoundException()
        previous_resolved = report.resolved

        new_report_comment = ReportComment(
            comment=report_comment,
            report_id=report_id,
            user_id=moderator.id,
        )
        db.session.add(new_report_comment)

        now = datetime.now(timezone.utc)
        report.updated_at = now
        report_action = None
        if resolved is not None:
            report.resolved = resolved
        if resolved is True and report.resolved_by is None:
            report.resolved_at = now
            report.resolved_by = moderator.id
            report_action = ReportAction(
                report_id=report.id,
                moderator_id=moderator.id,
                action_type="report_resolution",
                created_at=now,
            )
        if resolved is False:
            report.resolved_at = None
            report.resolved_by = None
            if previous_resolved is True:
                report_action = ReportAction(
                    report_id=report.id,
                    moderator_id=moderator.id,
                    action_type="report_reopening",
                    created_at=now,
                )

        if report_action:
            db.session.add(report_action)

        db.session.commit()

        return report

    @staticmethod
    def create_report_action(
        *,
        report: Report,
        moderator: User,
        action_type: str,
        reason: Optional[str] = None,
        data: Dict,
    ) -> Optional[ReportAction]:
        reported_user: User = report.reported_user

        # if reported user has been deleted after report creation
        if not reported_user:
            raise InvalidReportActionException("invalid 'username'")

        now = datetime.now(timezone.utc)
        report_action = None
        if action_type in ALL_USER_ACTION_TYPES:
            username = data.get("username")
            if not username:
                raise InvalidReportActionException("'username' is missing")
            if username != reported_user.username:
                raise InvalidReportActionException("invalid 'username'")

            if action_type.startswith("user_warning"):
                user = User.query.filter_by(username=username).first()
                if not user:
                    raise InvalidReportActionException("invalid 'username'")

                existing_report_action = ReportAction.query.filter_by(
                    action_type=action_type,
                    report_id=report.id,
                    user_id=user.id,
                ).first()
                if existing_report_action:
                    raise UserWarningExistsException("user already warned")

                report_action = ReportAction(
                    moderator_id=moderator.id,
                    action_type=action_type,
                    created_at=now,
                    report_id=report.id,
                    reason=reason,
                    user_id=user.id,
                )
                if report.reported_comment_id:
                    report_action.comment_id = report.reported_comment_id
                elif report.reported_workout_id:
                    report_action.workout_id = report.reported_workout_id
                db.session.add(report_action)
            else:
                user_manager_service = UserManagerService(
                    username=username, moderator_id=moderator.id
                )
                user, _, _, _ = user_manager_service.update(
                    suspended=action_type == "user_suspension",
                    report_id=report.id,
                    reason=reason,
                )
                if action_type == "user_unsuspension":
                    appeal = (
                        ReportActionAppeal.query.join(ReportAction)
                        .filter(
                            ReportAction.report_id == report.id,
                            ReportAction.user_id == user.id,
                            ReportAction.action_type == "user_suspension",
                        )
                        .first()
                    )
                    if appeal:
                        appeal.approved = None
                        appeal.updated_at = datetime.now(timezone.utc)
                        db.session.flush()

        elif action_type in COMMENT_ACTION_TYPES + WORKOUT_ACTION_TYPES:
            object_type = action_type.split("_")[0]
            object_type_column = f"{object_type}_id"
            object_id = data.get(object_type_column)
            if not object_id:
                raise InvalidReportActionException(
                    f"'{object_type_column}' is missing"
                )
            reported_object: Union[Comment, Workout] = getattr(
                report, f"reported_{object_type}"
            )
            if not reported_object or reported_object.short_id != object_id:
                raise InvalidReportActionException(
                    f"invalid '{object_type_column}'"
                )

            report_action = ReportAction(
                moderator_id=moderator.id,
                action_type=action_type,
                created_at=now,
                report_id=report.id,
                reason=reason,
                user_id=reported_object.user_id,
                **{object_type_column: reported_object.id},
            )
            db.session.add(report_action)

            if "_suspension" in action_type:
                if reported_object.suspended_at:
                    raise InvalidReportActionException(
                        f"{object_type} already suspended"
                    )
                reported_object.suspended_at = now

            else:
                if reported_object.suspended_at is None:
                    raise InvalidReportActionException(
                        f"{object_type} already reactivated"
                    )
                reported_object.suspended_at = None
                appeal = (
                    ReportActionAppeal.query.join(ReportAction)
                    .filter(
                        ReportAction.report_id == report.id,
                        ReportAction.user_id == reported_object.user_id,
                        ReportAction.action_type
                        == f"{object_type}_suspension",
                    )
                    .first()
                )
                if appeal:
                    appeal.approved = None
                    appeal.updated_at = datetime.now(timezone.utc)
                    db.session.flush()
            db.session.flush()
        else:
            raise InvalidReportActionException("invalid action type")
        return report_action

    @staticmethod
    def process_appeal(
        appeal: ReportActionAppeal, moderator: User, data: Dict
    ) -> Optional[ReportAction]:
        appeal.moderator_id = moderator.id
        appeal.approved = data["approved"]
        appeal.reason = data["reason"]
        appeal.updated_at = datetime.now(timezone.utc)

        action = appeal.action
        content = None
        content_type = ""
        if action.action_type.startswith("comment_"):
            content = Comment.query.filter_by(id=action.comment_id).first()
            content_type = "comment"
        elif action.action_type.startswith("workout_"):
            content = Workout.query.filter_by(id=action.workout_id).first()
            content_type = "workout"

        new_report_action = None
        if data["approved"]:
            if action.action_type == "user_suspension":
                if not appeal.user.suspended_at:
                    raise InvalidReportActionException(
                        "user account has already been reactivated"
                    )

                user_manager_service = UserManagerService(
                    username=appeal.user.username,
                    moderator_id=moderator.id,
                )
                user, _, _, new_report_action = user_manager_service.update(
                    suspended=False, report_id=appeal.action.report_id
                )
            if action.action_type == "user_warning":
                new_report_action = ReportAction(
                    moderator_id=moderator.id,
                    action_type="user_warning_lifting",
                    created_at=datetime.now(timezone.utc),
                    report_id=action.report_id,
                    user_id=action.user_id,
                )
                db.session.add(new_report_action)
            if (
                action.action_type
                in ["comment_suspension", "workout_suspension"]
                and content
            ):
                if not content.suspended_at:
                    raise InvalidReportActionException(
                        f"{content_type} already reactivated"
                    )
                content_id = {f"{content_type}_id": content.id}
                new_report_action = ReportAction(
                    moderator_id=moderator.id,
                    action_type=f"{content_type}_unsuspension",
                    created_at=datetime.now(timezone.utc),
                    report_id=action.report_id,
                    user_id=action.user_id,
                    **content_id,
                )
                db.session.add(new_report_action)
                content.suspended_at = None
        else:
            if (
                action.action_type == "user_suspension"
                and not appeal.user.suspended_at
            ):
                if not appeal.user.suspended_at:
                    raise InvalidReportActionException(
                        "user account has been reactivated after appeal"
                    )
            if (
                action.action_type
                in ["comment_suspension", "workout_suspension"]
                and content
                and not content.suspended_at
            ):
                raise InvalidReportActionException(
                    f"{content_type} has been reactivated after appeal"
                )
        return new_report_action



LANGUAGES_DATE_STRING = {
    "bg": "d MMM yyyy",  # Bulgarian
    # "bn": "d MMM, yyyy",  # Bengali  # disabled for now
    "cs": "d. MMM yyyy",  # Czech
    "de": "d. MMM yyyy",  # German
    "en": "MMM. d, yyyy",  # English
    "es": "d MMM yyyy",  # Spanish
    "eu": "yyyy MMM d",  # Basque
    "fr": "d MMM yyyy",  # French
    "gl": "d MMM yyyy",  # Galician
    "hr": "d. MMM yyyy.",  # Croatian
    "it": "d MMM yyyy",  # Italian
    "nb": "d. MMM yyyy",  # Norwegian Bokmål
    "nl": "d MMM. yyyy",  # Dutch
    "pl": "d MMM yyyy",  # Polish
    "pt": "d MMM yyyy",  # Portuguese
    "ru": "d MMM yyyy",  # Russian
    "zh_Hans": "y年M月d日",  # Chinese (Simplified Han script)
}

SUPPORTED_LANGUAGES = list(LANGUAGES_DATE_STRING.keys())



"""add notifications preferences

Revision ID: d3e40c2bda80
Revises: 8a80bec0a410
Create Date: 2025-01-01 18:28:15.250710

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'd3e40c2bda80'
down_revision = '8a80bec0a410'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('notification_preferences', postgresql.JSONB(astext_type=sa.Text()), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.drop_column('notification_preferences')

    # ### end Alembic commands ###



from datetime import datetime, timedelta, timezone
from typing import Optional

import jwt
from flask import current_app

from fittrackee.utils import clean


def get_user_token(
    user_id: int, password_reset: Optional[bool] = False
) -> str:
    """
    Return authentication token for a given user.
    Token expiration time depends on token type (authentication or password
    reset)
    """
    expiration_days: float = (
        0.0 if password_reset else current_app.config["TOKEN_EXPIRATION_DAYS"]
    )
    expiration_seconds: float = (
        current_app.config["PASSWORD_TOKEN_EXPIRATION_SECONDS"]
        if password_reset
        else current_app.config["TOKEN_EXPIRATION_SECONDS"]
    )
    now = datetime.now(timezone.utc)
    payload = {
        "exp": now
        + timedelta(days=expiration_days, seconds=expiration_seconds),
        "iat": now,
        "sub": str(user_id),
    }
    return jwt.encode(
        payload,
        current_app.config["SECRET_KEY"],
        algorithm="HS256",
    )


def decode_user_token(auth_token: str) -> int:
    """
    Return user id from token
    """
    payload = jwt.decode(
        auth_token,
        current_app.config["SECRET_KEY"],
        algorithms=["HS256"],
    )
    return int(payload["sub"])


def clean_blacklisted_tokens(days: int) -> int:
    """
    Delete blacklisted tokens expired for more than provided number of days
    """
    sql = """
        DELETE FROM blacklisted_tokens
        WHERE blacklisted_tokens.expired_at < :limit;
    """
    return clean(sql, days)



import hashlib
import random
from typing import Dict, List

from flask import current_app
from staticmap import Line, StaticMap

from fittrackee import VERSION
from fittrackee.files import get_absolute_file_path


def get_static_map_tile_server_url(tile_server_config: Dict) -> str:
    if tile_server_config["STATICMAP_SUBDOMAINS"]:
        subdomains = tile_server_config["STATICMAP_SUBDOMAINS"].split(",")
        subdomain = f"{random.choice(subdomains)}."  # nosec
    else:
        subdomain = ""
    return tile_server_config["URL"].replace("{s}.", subdomain)


def generate_map(map_filepath: str, map_data: List) -> None:
    """
    Generate and save map image from map data
    """
    m = StaticMap(
        width=400,
        height=225,
        padding_x=10,
        headers={"User-Agent": f"FitTrackee v{VERSION}"},
        delay_between_retries=5,
    )
    if not current_app.config["TILE_SERVER"]["DEFAULT_STATICMAP"]:
        m.url_template = get_static_map_tile_server_url(
            current_app.config["TILE_SERVER"]
        )
    line = Line(map_data, "#3388FF", 4)
    m.add_line(line)
    image = m.render()
    image.save(map_filepath)


def get_map_hash(map_filepath: str) -> str:
    """
    Generate a md5 hash used as id instead of workout id, to retrieve map
    image (maps are sensitive data)
    """
    md5 = hashlib.md5(usedforsecurity=False)
    absolute_map_filepath = get_absolute_file_path(map_filepath)
    with open(absolute_map_filepath, "rb") as f:
        for chunk in iter(lambda: f.read(128 * md5.block_size), b""):
            md5.update(chunk)
    return md5.hexdigest()



from fittrackee import dramatiq
from fittrackee.users.export_data import export_user_data


@dramatiq.actor(queue_name="fittrackee_users_exports")
def export_data(export_request_id: int) -> None:
    export_user_data(export_request_id)
